// BenLincoln.DaVinci - A shared library used by The Mirror's Surface Breaks
// Copyright 2010-2011 Ben Lincoln
// http://www.beneaththewaves.net/
//

// This file is part of BenLincoln.DaVinci.

// BenLincoln.DaVinci is free software: you can redistribute it and/or modify
// it under the terms of version 3 of the GNU General Public License as published by
// the Free Software Foundation.

// BenLincoln.DaVinci is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with BenLincoln.DaVinci (in the file LICENSE.txt).  
// If not, see <http://www.gnu.org/licenses/>.

using System;
using System.Collections.Generic;
using System.IO;
using System.Text;
using System.Text.RegularExpressions;

namespace BenLincoln.DaVinci
{
    public class DCSImage : MultiChannelImage
    {
        protected DCSParameterSet _DCSConfiguration;

        protected MultiChannelImage _SourceImage;

        // the script code used to generate the tinted version(s) of a channel
        //protected string _GenerateScriptCode;

        #region Properties

        public DCSParameterSet DCSConfiguration
        {
            get
            {
                return _DCSConfiguration;
            }
            set
            {
                _DCSConfiguration = value;
                _DCSConfiguration.NameRelatedValueChanged += new NameRelatedValueChangedHandler(DCSNameEventHander);
            }
        }

        public MultiChannelImage SourceImage
        {
            get
            {
                return _SourceImage;
            }
            set
            {
                _SourceImage = value;
            }
        }

        //public string GenerateScriptCode
        //{
        //    get
        //    {
        //        return _GenerateScriptCode;
        //    }
        //    set
        //    {
        //        _GenerateScriptCode = value;
        //    }
        //}

        #endregion

        public DCSImage(MultiChannelImage sourceImage, DCSParameterSet dcsConfig)
            : base()
        {
            _NamePrefix = "FC-DCS-";
            //_GenerateScriptCode = " = dcsf(todouble(%IMAGE%))";
            _SourceImage = sourceImage;
            _Channels = _SourceImage.Channels;
            DCSConfiguration = dcsConfig;
            UseInStandardLumaColourImages = dcsConfig.UseInStandardLumaColourImages;
            GenerateName();
        }

        public DCSImage(MultiChannelImage sourceImage)
            : this(sourceImage, new DCSParameterSet())
        {
        }

        protected void DCSNameEventHander(object sender, NameRelatedValueChangedArgs e)
        {
            GenerateName();
        }

        protected override void _GenerateName()
        {
            StringBuilder nameBuilder = new StringBuilder(_NamePrefix);

            string siName = _SourceImage.ImageName.Trim().Replace(" ", "_");
            nameBuilder.Append("(");
            nameBuilder.Append(siName);
            nameBuilder.Append(")");
            nameBuilder.Append(_NameComponentSeparator);
            nameBuilder.Append(GetDCSTypeName());
            nameBuilder.Append(_NameSuffix);
            _ImageName = nameBuilder.ToString();
        }

        public string GetDCSTypeName()
        {
            StringBuilder builder = new StringBuilder();
            bool appendPercent = false;
            switch (DCSConfiguration.Mode)
            {
                case DCSMode.Standard:
                    builder.Append("DCS_");
                    break;
                case DCSMode.Running:
                    builder.Append("RDCS_");
                    appendPercent = true;
                    break;
                default:
                    throw new ArgumentException("The DCS Type of this image ('" + _ImageName + "') is set to an unsupported value ('" + DCSConfiguration.Mode.ToString() + "')");
                    break;
            }
            switch (DCSConfiguration.ColourSpace)
            {
                case DCSColourSpace.RGB:
                    builder.Append("(RGB");
                    break;
                case DCSColourSpace.YUV:
                    builder.Append("(YUV");
                    break;
                case DCSColourSpace.HSL:
                    builder.Append("(HSL");
                    break;
                case DCSColourSpace.HSV:
                    builder.Append("(HSV");
                    break;
                case DCSColourSpace.HSY:
                    builder.Append("(HSY");
                    break;
                case DCSColourSpace.IHSL:
                    builder.Append("(IHSL");
                    break;
                default:
                    throw new ArgumentException("The colourspace of this image ('" + _ImageName + "') is set to an unsupported value ('" + DCSConfiguration.ColourSpace.ToString() + "')");
                    break;
            }
            if (appendPercent)
            {
                builder.Append(String.Format("-{0:00}", (DCSConfiguration.RDCSSectionPercent * 100.0f)));
                builder.Append("_Percent");
            }
            if (_DCSConfiguration.Ignore != DCSParameterSet.DefaultIgnore)
            {
                string ignoreVal = String.Format("{0:F3}", _DCSConfiguration.Ignore);
                if ((_DCSConfiguration.Ignore == 0) || (Math.Abs(_DCSConfiguration.Ignore) > 10))
                {
                    ignoreVal = String.Format("{0:F0}", _DCSConfiguration.Ignore);
                }
                ignoreVal = ignoreVal.Replace('.', 'd');
                builder.Append("-Ignore_");
                builder.Append(ignoreVal);
            }
            if (_DCSConfiguration.Variance != DCSParameterSet.DefaultVariance)
            {
                string varVal = String.Format("{0:F3}", _DCSConfiguration.Variance);
                if ((_DCSConfiguration.Variance == 0) || (Math.Abs(_DCSConfiguration.Variance) > 10))
                {
                    varVal = String.Format("{0:F0}", _DCSConfiguration.Variance);
                }
                varVal = varVal.Replace('.', 'd');
                builder.Append("-Variance_");
                builder.Append(varVal);
            }

            builder.Append(")");
            return builder.ToString();
        }

        public override string GetCreationScriptCode(string scriptTargetObjectName, bool regenerateExistingFiles)
        {
            StringBuilder builder = new StringBuilder();
            string filePath = GetFilePathWithoutExtension();

            if ((!regenerateExistingFiles) && AllOutputFilesExist(filePath, OutputFormats))
            {
                builder.Append("# Information: all file-type variations of the decorrelation stretch false colour image '");
                builder.Append(_ImageName);
                builder.Append("' already appear to exist. The option to regenerate existing files is not enabled, so the script will not overwrite them.");
                builder.Append(Environment.NewLine);
                return builder.ToString();
            }
            _WillBeCreatedByScript = true;

            builder.Append("# Create the '");
            builder.Append(ImageName);
            builder.Append("' decorrelation stretch image");
            builder.Append(Environment.NewLine);
            builder.Append("# Generate a decorrelation stretch variation of '");
            builder.Append(filePath);
            builder.Append("'.");
            builder.Append(Environment.NewLine);

            if ((regenerateExistingFiles) && (AnyOutputFilesExist(filePath, OutputFormats)))
            {
                builder.Append("# Warning: one or more variations of the target file already exist, but will be overwritten as the option to regenerate existing files is enabled.");
                builder.Append(Environment.NewLine);
                _WillOverwriteExistingData = true;
            }

            string sourceObjectName = _TemporaryObjectNamePrefix + "dcs";

            builder.Append(SourceImage.GetReadScriptCode(sourceObjectName));

            builder.Append(GetDCSScriptCode(scriptTargetObjectName, sourceObjectName));

            // normalize the image if that option is enabled
            //builder.Append(WrapScriptObjectInNormalizationCode(scriptTargetObjectName, OutputNormalization));
            //builder.Append(GetFullNormalizationWrappedLine(scriptTargetObjectName, OutputNormalization));
            //builder.Append(OutputTransformations.GetWrappedLine(scriptTargetObjectName));
            
            // write the image to disk
            builder.Append(GetImageFileWriteCode(scriptTargetObjectName, OutputFormats, filePath, regenerateExistingFiles));

            // clear out the temporary objects
            builder.Append(GreyscaleChannel.GetDeallocationScriptCode(sourceObjectName));
            builder.Append(GreyscaleChannel.GetDeallocationScriptCode(scriptTargetObjectName));
            //builder.Append(sourceObjectName);
            //builder.Append(_DeallocateObjectScriptCode);
            //builder.Append(Environment.NewLine);

            return builder.ToString();
        }

        protected string GetDCSScriptCode(string scriptTargetObjectName, string sourceObjectName)
        {
            string dcsCommand = "";

            switch (DCSConfiguration.ColourSpace)
            {
                case DCSColourSpace.RGB:
                    switch (DCSConfiguration.Mode)
                    {
                        case DCSMode.Standard:
                            dcsCommand = "%TARGETOBJECT% = dcsf(tofloat(%SOURCEOBJECT%), ignore = %IGNORE%, variance = %VARIANCE%)";
                            break;
                        case DCSMode.Running:
                            dcsCommand = "%TARGETOBJECT% = rdcsf(tofloat(%SOURCEOBJECT%), xsizepercent = %RDCSSEECTIONPERCENT%, ysizepercent = %RDCSSEECTIONPERCENT%, ignore = %IGNORE%, variance = %VARIANCE%)";
                            break;
                        default:
                            throw new ArgumentException("The DCS Type of this image ('" + _ImageName + "') is set to an unsupported value ('" + DCSConfiguration.Mode.ToString() + "')");
                            break;
                    }
                    break;
                case DCSColourSpace.YUV:
                    switch (DCSConfiguration.Mode)
                    {
                        case DCSMode.Standard:
                            dcsCommand = "%TARGETOBJECT% = ynuv2rgb(dcsf(rgb2ynuv(tofloat(%SOURCEOBJECT%)), ignore = %IGNORE%, variance = %VARIANCE%))";
                            break;
                        case DCSMode.Running:
                            dcsCommand = "%TARGETOBJECT% = ynuv2rgb(rdcsf(rgb2ynuv(tofloat(%SOURCEOBJECT%)), xsizepercent = %RDCSSEECTIONPERCENT%, ysizepercent = %RDCSSEECTIONPERCENT%, ignore = %IGNORE%, variance = %VARIANCE%))";
                            break;
                        default:
                            throw new ArgumentException("The DCS Type of this image ('" + _ImageName + "') is set to an unsupported value ('" + DCSConfiguration.Mode.ToString() + "')");
                            break;
                    }
                    break;
                case DCSColourSpace.HSL:
                    switch (DCSConfiguration.Mode)
                    {
                        case DCSMode.Standard:
                            dcsCommand = "%TARGETOBJECT% = hsl2rgb(dcsf(rgb2hsl(tofloat(%SOURCEOBJECT%)), ignore = %IGNORE%, variance = %VARIANCE%))";
                            break;
                        case DCSMode.Running:
                            dcsCommand = "%TARGETOBJECT% = hsl2rgb(rdcsf(rgb2hsl(tofloat(%SOURCEOBJECT%)), xsizepercent = %RDCSSEECTIONPERCENT%, ysizepercent = %RDCSSEECTIONPERCENT%, ignore = %IGNORE%, variance = %VARIANCE%))";
                            break;
                        default:
                            throw new ArgumentException("The DCS Type of this image ('" + _ImageName + "') is set to an unsupported value ('" + DCSConfiguration.Mode.ToString() + "')");
                            break;
                    }
                    break;
                case DCSColourSpace.HSV:
                    switch (DCSConfiguration.Mode)
                    {
                        case DCSMode.Standard:
                            dcsCommand = "%TARGETOBJECT% = hsv2rgb(dcsf(rgb2hsv(tofloat(%SOURCEOBJECT%)), ignore = %IGNORE%, variance = %VARIANCE%))";
                            break;
                        case DCSMode.Running:
                            dcsCommand = "%TARGETOBJECT% = hsv2rgb(rdcsf(rgb2hsv(tofloat(%SOURCEOBJECT%)), xsizepercent = %RDCSSEECTIONPERCENT%, ysizepercent = %RDCSSEECTIONPERCENT%, ignore = %IGNORE%, variance = %VARIANCE%))";
                            break;
                        default:
                            throw new ArgumentException("The DCS Type of this image ('" + _ImageName + "') is set to an unsupported value ('" + DCSConfiguration.Mode.ToString() + "')");
                            break;
                    }
                    break;
                case DCSColourSpace.HSY:
                    switch (DCSConfiguration.Mode)
                    {
                        case DCSMode.Standard:
                            dcsCommand = "%TARGETOBJECT% = hsy2rgb(dcsf(rgb2hsy(tofloat(%SOURCEOBJECT%)), ignore = %IGNORE%, variance = %VARIANCE%))";
                            break;
                        case DCSMode.Running:
                            dcsCommand = "%TARGETOBJECT% = hsy2rgb(rdcsf(rgb2hsy(tofloat(%SOURCEOBJECT%)), xsizepercent = %RDCSSEECTIONPERCENT%, ysizepercent = %RDCSSEECTIONPERCENT%, ignore = %IGNORE%, variance = %VARIANCE%))";
                            break;
                        default:
                            throw new ArgumentException("The DCS Type of this image ('" + _ImageName + "') is set to an unsupported value ('" + DCSConfiguration.Mode.ToString() + "')");
                            break;
                    }
                    break;
                case DCSColourSpace.IHSL:
                    switch (DCSConfiguration.Mode)
                    {
                        case DCSMode.Standard:
                            dcsCommand = "%TARGETOBJECT% = ihsl2rgb(dcsf(rgb2ihsl(tofloat(%SOURCEOBJECT%)), ignore = %IGNORE%, variance = %VARIANCE%))";
                            break;
                        case DCSMode.Running:
                            dcsCommand = "%TARGETOBJECT% = ihsl2rgb(rdcsf(rgb2ihsl(tofloat(%SOURCEOBJECT%)), xsizepercent = %RDCSSEECTIONPERCENT%, ysizepercent = %RDCSSEECTIONPERCENT%, ignore = %IGNORE%, variance = %VARIANCE%))";
                            break;
                        default:
                            throw new ArgumentException("The DCS Type of this image ('" + _ImageName + "') is set to an unsupported value ('" + DCSConfiguration.Mode.ToString() + "')");
                            break;
                    }
                    break;
                default:
                    throw new ArgumentException("The DCS colourspace of this image ('" + _ImageName + "') is set to an unsupported value ('" + DCSConfiguration.ColourSpace.ToString() + "')");
                    break;
            }

            dcsCommand = dcsCommand.Replace("%TARGETOBJECT%", scriptTargetObjectName);
            dcsCommand = dcsCommand.Replace("%SOURCEOBJECT%", sourceObjectName);
            dcsCommand = dcsCommand.Replace("%RDCSSEECTIONPERCENT%", DCSConfiguration.RDCSSectionPercent.ToString());
            dcsCommand = dcsCommand.Replace("%IGNORE%", DCSConfiguration.Ignore.ToString());
            dcsCommand = dcsCommand.Replace("%VARIANCE%", DCSConfiguration.Variance.ToString());

            StringBuilder builder = new StringBuilder(dcsCommand);

            builder.Append(Environment.NewLine);

            return builder.ToString();
        }

        //public override int GetMaximumFileCount()
        //{
        //    if (DCSConfiguration.GenerateDCSEnhancedContrast)
        //    {
        //        return OutputFormats.Length * 2;
        //    }
        //    return OutputFormats.Length;
        //}

        public override bool ContainsDisallowedChannels()
        {
            //foreach (GreyscaleChannel channel in Channels)
            //{
            //    if (!channel.UseAsThreeChannelChannel())
            //    {
            //        return true;
            //    }
            //}
            return false;
        }
    }
}
