// BenLincoln.DaVinci - A shared library used by The Mirror's Surface Breaks
// Copyright 2010-2011 Ben Lincoln
// http://www.beneaththewaves.net/
//

// This file is part of BenLincoln.DaVinci.

// BenLincoln.DaVinci is free software: you can redistribute it and/or modify
// it under the terms of version 3 of the GNU General Public License as published by
// the Free Software Foundation.

// BenLincoln.DaVinci is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with BenLincoln.DaVinci (in the file LICENSE.txt).  
// If not, see <http://www.gnu.org/licenses/>.

using System;
using System.Collections;
using System.Collections.Generic;
using System.IO;
using System.Text;
using System.Text.RegularExpressions;

namespace BenLincoln.DaVinci
{
    [Flags]
    public enum BandCombineMethod
    {
        None = 0,
        Minimum = 1,
        Maximum = 2,
        Average = 4,
        AverageDeviation = 8,
        StandardDeviation = 16,
        Variance = 32,
        Skewness = 64,
        Kurtosis = 128,
        Median = 256,
        Range = 512,
        FarthestFromAverageMin = 1024,
        FarthestFromAverageMax = 2048,
        FarthestFromMedianMin = 4096,
        FarthestFromMedianMax = 8192,
        ClosestToAverageMin = 16384,
        ClosestToAverageMax = 32768,
        Ratio = 524288,
        Difference = 1048576,
        NormalizedDifference = 2097152,
        NormalizedDifferenceCenteredOnZeroPointFive = 4194304,
        Luminance = 536870912,
        Custom = 1073741824
    }

    /// <summary>
    /// Represents a single greyscale channel (AKA "plane") of an image. Composed of 
    /// one or more spectral bands.
    /// </summary>
    public class GreyscaleChannel : OutputImage, IComparable
    {
        //protected string _ChannelName;

        //protected string _NamePrefix;

        protected SpectralBand[] _SpectralBands;
        protected BandCombineMethod _BandCombineMode;

        //protected OutputParameterSet[] _OutputFormats;

        protected bool _UseSecondaryImage;
        //protected bool RegenerateExistingImages;
        //protected bool _WillOverwriteExistingData;
        //protected bool _WillBeCreatedByScript;
        //protected Hashtable _TintingColours;

        protected bool _UseAsThreeChannelChannel;

        protected bool _ExplicitlySetUseAsThreeChannelChannel;

        protected bool _UsesAllSpectralBands;
        //protected bool _RenameAllBandsFiles;

        protected const float _LuminanceWeightRed = 0.299f;
        protected const float _LuminanceWeightBlue = 0.114f;

        protected string _CustomBandCombinationScriptCode;

        // for use in custom channels that reference other, previously-created channels
        protected GreyscaleChannel[] _ReferencedChannels;

        protected string _CustomDescription;

        protected const string _TemporaryObjectNamePrefix = "gsc_temp_";
        // which object type to have DaVinci cast to for the generated files
        // Note: this can be set to an empty string, but is likely to produce inconsistent results.
        //protected const string _InternalProcessingCast = "short";
        //protected const string _InternalProcessingCast = "tofloat";
        
        // which data organization mode to have DaVinci use when reading in and writing files
        //protected const string _DataOrganizationMode = "bip";
        // the script code to insert when an object should be removed from memory
        //protected const string _DeallocateObjectScriptCode = " = 0";


        //protected const string _ChannelFileOutputDirectory = "";
        //protected const string _ChannelFileExtension = ".TIF";
        //protected const string _ChannelFileDVType = "tiff";
        //protected const string _ChannelFileDVOverwrite = "1";

        #region Properties

        //public string ChannelName
        //{
        //    get
        //    {
        //        return _ChannelName;
        //    }
        //    set
        //    {
        //        _ChannelName = value;
        //    }
        //}

        //public string NamePrefix
        //{
        //    get
        //    {
        //        return _NamePrefix;
        //    }
        //    set
        //    {
        //        _NamePrefix = value;
        //    }
        //}

        public SpectralBand[] SpectralBands
        {
            get
            {
                return _SpectralBands;
            }
            set
            {
                _SpectralBands = value;
                GenerateName();
                if (!_ExplicitlySetUseAsThreeChannelChannel)
                {
                    _UseAsThreeChannelChannel = true;
                    foreach (SpectralBand band in SpectralBands)
                    {
                        if (band != null)
                        {
                            if (!band.UseAsThreeChannelChannel)
                            {
                                _UseAsThreeChannelChannel = false;
                            }
                        }
                    }
                }
            }
        }

        /// <summary>
        /// If this value is true, then the secondary version of the base spectral band 
        /// will be returned. This option is only used if there is only one spectral band
        /// in the channel.
        /// </summary>
        public bool UseSecondaryImage
        {
            get
            {
                return _UseSecondaryImage;
            }
            set
            {
                _UseSecondaryImage = value;
            }
        }

        /// <summary>
        /// If this value is true, then any multi-band greyscale channel files will be 
        /// overwritten as if they did not already exist. If this value is false, then
        /// the generated script will not recreate them. Defaults to false to avoid 
        /// unnecessary processing.
        /// </summary>
        //public bool regenerateExistingFiles
        //{
        //    get
        //    {
        //        return RegenerateExistingImages;
        //    }
        //    set
        //    {
        //        RegenerateExistingImages = value;
        //    }
        //}

        ///// <summary>
        ///// This read-only value represents whether or not running the script which represents
        ///// the channel will overwrite data which already exists on disk.
        ///// </summary>
        //public bool WillOverwriteExistingData
        //{
        //    get
        //    {
        //        return _WillOverwriteExistingData;
        //    }
        //}

        ///// <summary>
        ///// This read-only value represents whether or not running the script which represents
        ///// the channel will create a file
        ///// </summary>
        //public bool WillBeCreatedByScript
        //{
        //    get
        //    {
        //        return _WillBeCreatedByScript;
        //    }
        //}

        public bool UseAsThreeChannelChannel
        {
            get
            {
                if (_ExplicitlySetUseAsThreeChannelChannel)
                {
                    return _UseAsThreeChannelChannel;
                }
                foreach (SpectralBand band in SpectralBands)
                {
                    if (!band.UseAsThreeChannelChannel)
                    {
                        return false;
                    }
                }
                return true;
            }
            set
            {
                _UseAsThreeChannelChannel = value;
                _ExplicitlySetUseAsThreeChannelChannel = true;
            }
        }

        //public bool UseAsThreeChannelChannel
        //{
        //    get
        //    {
        //        foreach (SpectralBand band in SpectralBands)
        //        {
        //            if (!band.UseAsThreeChannelChannel)
        //            {
        //                return false;
        //            }
        //        }
        //        return true;
        //    }
        //}

        /// <summary>
        /// The method used to determine the result when multiple bands are combined (min, max, average, etc.).
        /// </summary>
        public BandCombineMethod BandCombineMode
        {
            get
            {
                return _BandCombineMode;
            }
            set
            {
                _BandCombineMode = value;
                GenerateName();
            }
        }

        //public OutputParameterSet[] OutputFormats
        //{
        //    get
        //    {
        //        return _OutputFormats;
        //    }
        //    set
        //    {
        //        _OutputFormats = value;
        //    }
        //}

        public bool UsesAllSpectralBands
        {
            get
            {
                return _UsesAllSpectralBands;
            }
            set
            {
                _UsesAllSpectralBands = value;
            }
        }

        //public bool RenameAllBandsFiles
        //{
        //    get
        //    {
        //        return _RenameAllBandsFiles;
        //    }
        //    set
        //    {
        //        _RenameAllBandsFiles = value;
        //    }
        //}

        public string CustomBandCombinationScriptCode
        {
            get
            {
                return _CustomBandCombinationScriptCode;
            }
            set
            {
                _CustomBandCombinationScriptCode = value;
                //if (value != "")
                //{
                //    _NamePrefix = "GS-Custom-";
                //}
            }
        }

        public GreyscaleChannel[] ReferencedChannels
        {
            get
            {
                return _ReferencedChannels;
            }
            set
            {
                _ReferencedChannels = value;
            }
        }

        public string CustomDescription
        {
            get
            {
                return _CustomDescription;
            }
            set
            {
                _CustomDescription = value;
            }
        }

        #endregion

        //public GreyscaleChannel(string name, int numBands, BandCombineMethod combineMode)
        public GreyscaleChannel(int numBands, BandCombineMethod combineMode)
            : base()
        {
            //_TintingColours = new Hashtable();
            //_ChannelName = name;
            _NamePrefix = "GS-";
            _SpectralBands = new SpectralBand[numBands];
            _ReferencedChannels = new GreyscaleChannel[0];
            _BandCombineMode = combineMode;
            _UseSecondaryImage = false;
            //RegenerateExistingImages = false;
            _UsesAllSpectralBands = false;
            //_RenameAllBandsFiles = false;
            _UseAsThreeChannelChannel = true;
            _ExplicitlySetUseAsThreeChannelChannel = false;
            CustomBandCombinationScriptCode = "";
            CustomDescription = "custom specification";
            //_MomentCubeWindowSize = 1;
        }

        public GreyscaleChannel(SpectralBand[] bands, BandCombineMethod combineMode)
            : this(bands.Length, combineMode)
        {
            _SpectralBands = bands;
            //_OutputFormats[0].BitsPerChannel = BitsPerChannel.SixteenShort;
        }

        public GreyscaleChannel()
            : this(0, BandCombineMethod.Average)
        {
        }

        protected override void _GenerateName()
        {
            if (SpectralBands.Length == 0)
            {
                _ImageName = "";
                return;
            }
            //StringBuilder nameBuilder = new StringBuilder(_NamePrefix);
            StringBuilder nameBuilder = new StringBuilder();
            //if ((UsesAllSpectralBands) && (RenameAllBandsFiles))
            //if ((UsesAllSpectralBands) && (RenameAllBandsFiles))
            //{
            //    nameBuilder.Append("All_Bands");
            //}
            //else
            //{
                for (int bandNum = 0; bandNum < SpectralBands.Length; bandNum++)
                {
                    if (SpectralBands[bandNum] == null)
                    {
                        nameBuilder.Append("Null");
                    }
                    else
                    {
                        nameBuilder.Append(SpectralBands[bandNum].BandName);
                    }
                    if (bandNum < (SpectralBands.Length - 1))
                    {
                        nameBuilder.Append("_");
                    }
                }
            //}
            if (SpectralBands.Length > 1)
            {
                nameBuilder.Append(GetBandCombineModeNameSuffix(BandCombineMode));
            }

            nameBuilder.Append(_NameSuffix);
            _ImageName = nameBuilder.ToString();
        }

        public static string GetBandCombineModeNameSuffix(BandCombineMethod mode)
        {
            return "_(" + GetBandCombineMethodNameString(mode) + ")";
        }

        /// <summary>
        /// Gets the DaVinci script code to read the channel in from a file.
        /// If the file is composed of more than one spectral band, this assumes the
        /// combined file has been created.
        /// </summary>
        /// <param name="scriptTargetObjectName">The script object name which should be populated with the channel.</param>
        /// <returns></returns>
        public string GetChannelFileReadScriptCode(string scriptPrepend, string scriptTargetObjectName)
        {
            StringBuilder builder = new StringBuilder(scriptPrepend);
            builder.Append("# Read in the channel '");
            builder.Append(ImageName);
            builder.Append("' and place its contents into the object '");
            builder.Append(scriptTargetObjectName);
            builder.Append("'.");
            builder.Append(Environment.NewLine);
            switch (_SpectralBands.Length)
            {
                case 0:
                    throw new ArgumentOutOfRangeException("The channel '" + ImageName + "' has no source spectral bands");
                    break;
                case 1:
                    return GetSingleFileReadScriptCode(scriptTargetObjectName);
                    break;
                default:
                    return GetMultiFileReadScriptCode(scriptTargetObjectName);
                    break;
            }
        }

        protected override string _GetReadScriptCode(string scriptPrepend, string scriptTargetObjectName)
        {
            return GetChannelFileReadScriptCode(scriptPrepend, scriptTargetObjectName);
        }

        public override string GetName()
        {
            //return ChannelName;
            return ImageName;
        }

        public bool IsSingleSpectralBand
        {
            get
            {
                if (_SpectralBands.Length == 1)
                {
                    return true;
                }
                return false;
            }
        }

        public bool HasSecondarySourceImage
        {
            get
            {
                if (IsSingleSpectralBand)
                {
                    if (_SpectralBands[0].HasSecondaryImage)
                    {
                        return true;
                    }
                }
                return false;
            }
        }

        protected string GetSingleFileReadScriptCode(string scriptTargetObjectName)
        {
            //StringBuilder builder = new StringBuilder(scriptTargetObjectName);
            StringBuilder builder = new StringBuilder();
            builder.Append(scriptTargetObjectName);
            builder.Append(" = ");
            //builder.Append(_CastOnLoad);
            //builder.Append("(");
            //builder.Append(_DataOrganizationMode);
            //builder.Append("(read(\"");
            //SpectralBandData sourceBandData = _SpectralBands[0].GetPrimaryImageData();
            if (_UseSecondaryImage)
            {
                //builder.Append(GetNormalizationCodePre(_SpectralBands[0].InputNormalization));
                builder.Append(_CastOnLoad);
                builder.Append("(");
                builder.Append(_DataOrganizationMode);
                builder.Append("(read(\"");

                SpectralBandData sourceBandData = _SpectralBands[0].GetSecondaryImageData();
                builder.Append(sourceBandData.FileName);

                builder.Append("\")))");
                //builder.Append(GetNormalizationCodePost(_SpectralBands[0].InputNormalization));
                builder.Append(Environment.NewLine);
                //builder.Append(scriptTargetObjectName);
                //builder.Append(" = ");
                //builder.Append(scriptTargetObjectName);
                //builder.Append(sourceBandData.Plane);
                string objectName = scriptTargetObjectName + sourceBandData.Plane;
                builder.Append(_SpectralBands[0].InputTransformations.GetWrappedLine(scriptTargetObjectName, objectName));
            }
            else
            {
                builder.Append(_CastOnLoad);
                builder.Append("(");
                builder.Append(_DataOrganizationMode);
                builder.Append("(read(\"");
                builder.Append(GetFilePathWithPrimaryExtension());
                builder.Append("\")))");
            }
            //builder.Append(sourceBandData.FileName);
            //builder.Append("\")))");
            //builder.Append(Environment.NewLine);
            //builder.Append(scriptTargetObjectName);
            //builder.Append(" = ");
            //builder.Append(scriptTargetObjectName);
            //builder.Append(sourceBandData.Plane);
            builder.Append(Environment.NewLine);
            string result = builder.ToString();
            //result = scriptTargetObjectName + " = " + 
            //    WrapScriptObjectInNormalizationCode(result, _SpectralBands[0].InputNormalization) + 
            //    Environment.NewLine;
            //builder.Append(Environment.NewLine);
            //return builder.ToString();
            return result;
        }

        protected string GetMultiFileReadScriptCode(string scriptTargetObjectName)
        {
            StringBuilder builder = new StringBuilder(scriptTargetObjectName);
            builder.Append(" = ");
            builder.Append(_CastOnLoad);
            builder.Append("(");
            builder.Append(_DataOrganizationMode);
            builder.Append("(read(\"");
            builder.Append(GetFilePathWithPrimaryExtension());
            builder.Append("\")))");
            builder.Append(Environment.NewLine);
            return builder.ToString();
        }

        /// <summary>
        /// Gets the DaVinci script code to create the object from its base components.
        /// </summary>
        /// <param name="scriptTargetObjectName">The script object name which should be populated with the channel.</param>
        /// <returns></returns>
        public override string GetCreationScriptCode(string scriptTargetObjectName, bool regenerateExistingFiles)
        {
            _FileCountWillBeWritten = 0;
            _FileCountWillBeSkipped = 0;

            GetCustomCodePrerequisites();

            if (((BandCombineMode & BandCombineMethod.Custom) == BandCombineMethod.Custom) && (CustomBandCombinationScriptCode != ""))
                {
                    return GetCustomMultiFileCreationScriptCode(scriptTargetObjectName, regenerateExistingFiles);
                }

            switch (_SpectralBands.Length)
            {
                case 0:
                    throw new ArgumentOutOfRangeException("The channel '" + ImageName + "' has no source spectral bands");
                    break;
                case 1:
                    return GetSingleFileCreationScriptCode(scriptTargetObjectName, regenerateExistingFiles);
                    break;
                default:
                    return GetMultiFileCreationScriptCode(scriptTargetObjectName, regenerateExistingFiles);
                    break;
            }
        }

        protected string GetCustomMultiFileCreationScriptCode(string scriptTargetObjectName, bool regenerateExistingFiles)
        {
            StringBuilder builder = new StringBuilder();

            if (AllOutputFilesExist(GetFilePathWithoutExtension(), OutputFormats))
            {
                _FileCountWillBeSkipped = OutputFormats.Length;
                builder.Append("# Information: all file-type variations of the custom greyscale channel image '");
                builder.Append(ImageName);
                builder.Append("' already appear to exist. The option to regenerate existing files is not enabled, so the script will not overwrite them.");
                builder.Append(Environment.NewLine);
                return builder.ToString();
            }

            builder.Append("# Create the '");
            builder.Append(ImageName);
            builder.Append("' greyscale channel image (");
            builder.Append(CustomDescription);
            builder.Append(")");
            builder.Append(Environment.NewLine);
            builder.Append("# Read in the prerequisite images.");
            builder.Append(Environment.NewLine);

            string customCreationCode = CustomBandCombinationScriptCode;

            CustomChannelReference[] cRefs = GetPrerequisiteReferences();

            foreach (CustomChannelReference cRef in cRefs)
            {
                customCreationCode = customCreationCode.Replace(cRef.ScriptTextToReplace, cRef.ScriptObjectName);
                GreyscaleChannel refChannel = GetChannelByName(cRef.ReferenceName);
                builder.Append(refChannel.GetChannelFileReadScriptCode("", cRef.ScriptObjectName));
                builder.Append(InputTransformations.GetWrappedLine(cRef.ScriptObjectName));
            }

            builder.Append("# Execute the custom creation command.");
            builder.Append(Environment.NewLine);

            builder.Append(scriptTargetObjectName);
            builder.Append(" ");
            builder.Append(customCreationCode);
            builder.Append(Environment.NewLine);

            //builder.Append(OutputTransformations.GetWrappedLine(scriptTargetObjectName));

            builder.Append(GetChannelFileWriteCode(scriptTargetObjectName, regenerateExistingFiles));

            builder.Append("# Deallocate temporary objects.");
            builder.Append(Environment.NewLine);

            foreach (CustomChannelReference cRef in cRefs)
            {
                builder.Append(GetDeallocationScriptCode(cRef.ScriptObjectName));
            }
            builder.Append(GetDeallocationScriptCode(scriptTargetObjectName));

            return builder.ToString();
        }

        //protected string GetCustomReferenceReadCode(CustomChannelReference r)
        //{
        //    switch (r.ReferenceType)
        //    {
        //        case CustomChannelReferenceType.BandByName:
        //            if (ContainsChannel(r.ReferenceName))
        //            {
        //                AddSpectralBand(cube.GetSpectralBand(r.ReferenceName));
        //                AddReferencedChannel(cube.GetGreyscaleChannel(r.ReferenceName));
        //            }
        //            else
        //            {
        //                throw new ArgumentException("Could not find a spectral band named '" + r.ReferenceName + "' in the image cube.");
        //            }
        //            break;
        //        case CustomChannelReferenceType.ChannelByName:
        //            if (cube.ContainsGreyscaleChannel(r.ReferenceName))
        //            {
        //                AddReferencedChannel(cube.GetGreyscaleChannel(r.ReferenceName));
        //            }
        //            else
        //            {
        //                throw new ArgumentException("Could not find a channel named '" + r.ReferenceName + "' in the image cube.");
        //            }
        //            break;
        //        case CustomChannelReferenceType.SpecialBand:
        //            throw new ArgumentException("This code should never be executed - these references should be replaced earlier in the channel's creation process.");
        //            break;
        //    }
        //}

        protected string GetSingleFileCreationScriptCode(string scriptTargetObjectName, bool regenerateExistingFiles)
        {
            // this code is only necessary if the source file doesn't already exist
            // note that this code cannot be overridden by the use of the RegenerateExistingImages
            // flag because it would overwrite the actual source files
            StringBuilder builder = new StringBuilder();
            if (AllOutputFilesExist(GetFilePathWithoutExtension(), OutputFormats))
            {
                _FileCountWillBeSkipped = OutputFormats.Length;
                builder.Append("# Information: all file-type variations of the greyscale channel image '");
                builder.Append(ImageName);
                builder.Append("' already appear to exist. The option to regenerate existing files is not enabled, so the script will not overwrite them.");
                builder.Append(Environment.NewLine);
                builder.Append("# Note that this is the standard behaviour when the name of an input file matches the name of an output file.");
                builder.Append(Environment.NewLine);
                builder.Append("printf(\"Skipping creation of output files for the image '" + ImageName + "' because all configured output files already exist.\\n\")");
                builder.Append(Environment.NewLine);
                return builder.ToString();                
            }

            _WillBeCreatedByScript = true;

            //_FileCountWillBeWritten = OutputFormats.Length;

            builder.Append("# Create the '");
            builder.Append(ImageName);
            builder.Append("' greyscale channel image (single source file)");
            builder.Append(Environment.NewLine);
            builder.Append(scriptTargetObjectName);
            builder.Append(" = ");

            StringBuilder b2 = new StringBuilder();

            //builder.Append(GetNormalizationCodePre(_SpectralBands[0].InputNormalization));
            b2.Append(_CastOnLoad);
            b2.Append("(");
            b2.Append(_DataOrganizationMode);
            b2.Append("(read(\"");
            SpectralBandData sourceBandData = _SpectralBands[0].GetPrimaryImageData();
            if (_UseSecondaryImage)
            {
                sourceBandData = _SpectralBands[0].GetSecondaryImageData();
            }
            b2.Append(sourceBandData.FileName);
            b2.Append("\")))");
            //builder.Append(GetNormalizationCodePost(_SpectralBands[0].InputNormalization));

            //builder.Append(WrapScriptObjectInNormalizationCode(b2.ToString(), _SpectralBands[0].InputNormalization));

            builder.Append(_SpectralBands[0].InputTransformations.WrapScriptObjectInTransformationCode(b2.ToString()));
            builder.Append(Environment.NewLine);

            builder.Append(InputTransformations.WrapScriptObjectInTransformationCode(b2.ToString()));
            builder.Append(Environment.NewLine);

            builder.Append(scriptTargetObjectName);
            builder.Append(" = ");
            builder.Append(scriptTargetObjectName);
            builder.Append(sourceBandData.Plane);
            builder.Append(Environment.NewLine);
            builder.Append(GetChannelFileWriteCode(scriptTargetObjectName, regenerateExistingFiles));
            //builder.Append(Environment.NewLine);
            //builder.Append(Environment.NewLine);

            // de-allocate objects
            builder.Append(GreyscaleChannel.GetDeallocationScriptCode(scriptTargetObjectName));
            return builder.ToString();
        }

        protected string GetMultiFileCreationScriptCode(string scriptTargetObjectName, bool regenerateExistingFiles)
        {
            return GetMultiFileCreationScriptCode(scriptTargetObjectName, regenerateExistingFiles, true, "", BandCombineMode, true, "", 1, null, null);
        }

        public static bool IncludesBandCombineMethod(BandCombineMethod methodSet, BandCombineMethod checkMethod)
        {
            if ((methodSet & checkMethod) == checkMethod)
            {
                return true;
            }
            return false;
        }

        public static bool RequiresMomentCubeCreation(BandCombineMethod allSetBandCombineMethods)
        {
            //BandCombineMethod asbcm = allSetBandCombineMethods ^ BandCombineMethod.Custom;
            //BandCombineMethod nonMomentMethods = (BandCombineMethod.Minimum | BandCombineMethod.Maximum | BandCombineMethod.Average 
            //    | BandCombineMethod.Ratio | BandCombineMethod.Difference 
            //    | BandCombineMethod.NormalizedDifference | BandCombineMethod.NormalizedDifferenceCenteredOnZeroPointFive 
            //    | BandCombineMethod.Luminance);
            //if ((asbcm ^ (nonMomentMethods)) > nonMomentMethods)
            //{
            //    return true;
            //}
            BandCombineMethod asbcm = allSetBandCombineMethods;
            if (IncludesBandCombineMethod(allSetBandCombineMethods, BandCombineMethod.Custom))
            {
                asbcm = allSetBandCombineMethods ^ BandCombineMethod.Custom;
            }
            
            if (IncludesBandCombineMethod(asbcm, BandCombineMethod.AverageDeviation))
            {
                return true;
            }
            if (IncludesBandCombineMethod(asbcm, BandCombineMethod.ClosestToAverageMax))
            {
                return true;
            }
            if (IncludesBandCombineMethod(asbcm, BandCombineMethod.ClosestToAverageMin))
            {
                return true;
            }
            if (IncludesBandCombineMethod(asbcm, BandCombineMethod.FarthestFromAverageMax))
            {
                return true;
            }
            if (IncludesBandCombineMethod(asbcm, BandCombineMethod.FarthestFromAverageMin))
            {
                return true;
            }
            if (IncludesBandCombineMethod(asbcm, BandCombineMethod.FarthestFromMedianMax))
            {
                return true;
            }
            if (IncludesBandCombineMethod(asbcm, BandCombineMethod.FarthestFromMedianMin))
            {
                return true;
            }
            if (IncludesBandCombineMethod(asbcm, BandCombineMethod.Kurtosis))
            {
                return true;
            }
            if (IncludesBandCombineMethod(asbcm, BandCombineMethod.Median))
            {
                return true;
            }
            if (IncludesBandCombineMethod(asbcm, BandCombineMethod.Range))
            {
                return true;
            }
            if (IncludesBandCombineMethod(asbcm, BandCombineMethod.Skewness))
            {
                return true;
            }
            if (IncludesBandCombineMethod(asbcm, BandCombineMethod.StandardDeviation))
            {
                return true;
            }
            if (IncludesBandCombineMethod(asbcm, BandCombineMethod.Variance))
            {
                return true;
            }

            return false;
        }

        public string GetMultiFileCreationScriptCode(string scriptTargetObjectName, bool regenerateExistingFiles, bool generateBandCube, string existingBandCubeName, BandCombineMethod allSetBandCombineMethods,
            bool generateMomentCube, string existingMomentCubeName, int windowSize, TransformationProfile cubeInputTransformations, TransformationProfile cubeOutputTransformations)
        {
            string filePath = GetFilePathWithoutExtension();
            StringBuilder builder = new StringBuilder();
            // if the option to recreate existing channnels is not enabled and the 
            // file already exists, do nothing.
            if ((!regenerateExistingFiles) && AllOutputFilesExist(filePath, OutputFormats))
            {
                _FileCountWillBeSkipped = OutputFormats.Length;
                builder.Append("# Information: all file-type variations of the greyscale channel image '");
                builder.Append(ImageName);
                builder.Append("' already appear to exist. The option to regenerate existing files is not enabled, so the script will not overwrite them.");
                builder.Append(Environment.NewLine);
                builder.Append("printf(\"Skipping creation of output files for the image '" + ImageName + "' because all configured output files already exist.\\n\")");
                builder.Append(Environment.NewLine);
                return builder.ToString();
            }
            //_FileCountWillBeWritten = OutputFormats.Length;

            _WillBeCreatedByScript = true;

            builder.Append("# Create the '");
            builder.Append(ImageName);
            builder.Append("' greyscale channel image (multiple source files)");
            builder.Append(Environment.NewLine);

            if ((regenerateExistingFiles) && (AnyOutputFilesExist(filePath, OutputFormats)))
            {
                builder.Append("# Warning: one or more variations of the target file already exist, but will be overwritten as the option to regenerate existing files is enabled.");
                builder.Append(Environment.NewLine);
                _WillOverwriteExistingData = true;
            }

            //// read in the files
            //for (int bandNum = 0; bandNum < _SpectralBands.Length; bandNum++)
            //{
            //    string tempBandName = _TemporaryObjectNamePrefix + String.Format("{0:000}", bandNum);
            //    builder.Append(tempBandName);
            //    builder.Append(" = ");
            //    if (_SpectralBands[bandNum].NormalizeInput)
            //    {
            //        builder.Append("normalize(");
            //    }
            //    builder.Append(_CastOnLoad);
            //    builder.Append("(");
            //    builder.Append(_DataOrganizationMode);
            //    builder.Append("(read(\"");
            //    builder.Append(_SpectralBands[bandNum].GetPrimaryImageData().FileName);
            //    builder.Append("\")))");
            //    if (_SpectralBands[bandNum].NormalizeInput)
            //    {
            //        builder.Append(")");
            //    }
            //    builder.Append(Environment.NewLine);
            //    builder.Append(tempBandName);
            //    builder.Append(" = ");
            //    builder.Append(tempBandName);
            //    builder.Append(_SpectralBands[bandNum].GetPrimaryImageData().Plane);
            //    builder.Append(Environment.NewLine);
            //}

            //// concatenate them together along the z-axis
            //// create the concatenation object and assign the first band to it
            //builder.Append(scriptTargetObjectName);
            //builder.Append(" = ");
            //builder.Append(_TemporaryObjectNamePrefix);
            //builder.Append("000");
            //builder.Append(Environment.NewLine);
            //// concatenate in the other band(s)
            //for (int bandNum = 1; bandNum < _SpectralBands.Length; bandNum++)
            //{
            //    builder.Append(scriptTargetObjectName);
            //    builder.Append(" = cat(");
            //    builder.Append(scriptTargetObjectName);
            //    builder.Append(", ");
            //    builder.Append(_TemporaryObjectNamePrefix);
            //    builder.Append(String.Format("{0:000}", bandNum));
            //    builder.Append(", axis = z)");
            //    builder.Append(Environment.NewLine);
            //}

            string bandCubeName = existingBandCubeName;

            if (generateBandCube)
            {
                bandCubeName = scriptTargetObjectName;
                builder.Append(GetBandCubeScriptCode(scriptTargetObjectName));
            }

            bool getAvgMinMaxFromMomentCube = false;
            bool mustGenerateMomentCube = false;
            // if only avg/min/max/luminance are being used, do not explicitly generate the moment cube, but do get them from the moment cube if one has already been created
            // (to avoid duplicate calculations)
            //if ((allSetBandCombineMethods ^ ((ulong)(BandCombineMethod.Minimum | BandCombineMethod.Maximum | BandCombineMethod.Average | BandCombineMethod.Luminance))) > 0)
            //{
            //    getAvgMinMaxFromMomentCube = true;
            //    if ((generateMomentCube) || (existingMomentCubeName == ""))
            //    {
            //        mustGenerateMomentCube = true;
            //    }
            //}
            if (RequiresMomentCubeCreation(allSetBandCombineMethods))
            {
                getAvgMinMaxFromMomentCube = true;
                if ((generateMomentCube) || (existingMomentCubeName == ""))
                {
                    mustGenerateMomentCube = true;
                }
            }

            string scriptMomentCubeName = existingMomentCubeName;

            if (mustGenerateMomentCube)
            {
                scriptMomentCubeName = _TemporaryObjectNamePrefix + "momentCube";
                builder.Append(GetMomentCubeScriptCode(scriptMomentCubeName, bandCubeName, allSetBandCombineMethods, windowSize, null, null));
            }

            // collapse the concatenated bands down to one plane via whichever
            // method is selected for this channel
            builder.Append(scriptTargetObjectName);
            builder.Append(" = ");

            //builder.Append(GetNormalizationCodePre(OutputNormalization));

            builder.Append(_CastOnLoad);
            //builder.Append(_DataOrganizationMode);
            builder.Append("(");
            builder.Append(_DataOrganizationMode);
            //builder.Append(_InternalProcessingCast);
            builder.Append("(");

            switch (_BandCombineMode)
            {
                case BandCombineMethod.Average:
                    if (getAvgMinMaxFromMomentCube)
                    {
                        builder.Append(scriptMomentCubeName);
                        builder.Append("[0,0,3]");
                    }
                    else
                    {
                        builder.Append("avg(");
                        builder.Append(bandCubeName);
                        builder.Append(", axis = z)");
                        //builder.Append(Environment.NewLine);
                    }
                    break;
                case BandCombineMethod.Minimum:
                    if (getAvgMinMaxFromMomentCube)
                    {
                        builder.Append(scriptMomentCubeName);
                        builder.Append("[0,0,1]");
                    }
                    else
                    {
                        builder.Append("min(");
                        builder.Append(bandCubeName);
                        builder.Append(", axis = z)");
                        //builder.Append(", axis = z)))");
                        //builder.Append(Environment.NewLine);
                    }
                    break;
                case BandCombineMethod.Maximum:
                    if (getAvgMinMaxFromMomentCube)
                    {
                        builder.Append(scriptMomentCubeName);
                        builder.Append("[0,0,2]");
                    }
                    else
                    {
                        builder.Append("max(");
                        builder.Append(bandCubeName);
                        builder.Append(", axis = z)");
                        //builder.Append(", axis = z)))");
                        //builder.Append(Environment.NewLine);
                    }
                    break;
                case BandCombineMethod.AverageDeviation:
                    builder.Append(scriptMomentCubeName);
                    builder.Append("[0,0,4]");
                    break;
                case BandCombineMethod.StandardDeviation:
                    builder.Append(scriptMomentCubeName);
                    builder.Append("[0,0,5]");
                    break;
                case BandCombineMethod.Variance:
                    builder.Append(scriptMomentCubeName);
                    builder.Append("[0,0,6]");
                    break;
                case BandCombineMethod.Skewness:
                    builder.Append(scriptMomentCubeName);
                    builder.Append("[0,0,7]");
                    break;
                case BandCombineMethod.Kurtosis:
                    builder.Append(scriptMomentCubeName);
                    builder.Append("[0,0,8]");
                    break;
                case BandCombineMethod.Median:
                    builder.Append(scriptMomentCubeName);
                    builder.Append("[0,0,9]");
                    break;
                case BandCombineMethod.Range:
                    builder.Append(scriptMomentCubeName);
                    builder.Append("[0,0,10]");
                    break;
                case BandCombineMethod.FarthestFromAverageMin:
                    builder.Append(scriptMomentCubeName);
                    builder.Append("[0,0,11]");
                    break;
                case BandCombineMethod.FarthestFromAverageMax:
                    builder.Append(scriptMomentCubeName);
                    builder.Append("[0,0,12]");
                    break;
                case BandCombineMethod.FarthestFromMedianMin:
                    builder.Append(scriptMomentCubeName);
                    builder.Append("[0,0,13]");
                    break;
                case BandCombineMethod.FarthestFromMedianMax:
                    builder.Append(scriptMomentCubeName);
                    builder.Append("[0,0,14]");
                    break;
                case BandCombineMethod.ClosestToAverageMin:
                    builder.Append(scriptMomentCubeName);
                    builder.Append("[0,0,15]");
                    break;
                case BandCombineMethod.ClosestToAverageMax:
                    builder.Append(scriptMomentCubeName);
                    builder.Append("[0,0,16]");
                    break;
                case BandCombineMethod.Luminance:
                    if (_SpectralBands.Length != 3)
                    {
                        throw new ArgumentException(String.Format("The Luminance combination mode is only valid for channels made up of exactly three spectral bands. This channel contains {D1} bands.", _SpectralBands.Length));
                    }
                    float weightGreen = 1.0f - _LuminanceWeightRed - _LuminanceWeightBlue;
                    if ((weightGreen < 0.0f) || (weightGreen > 1.0f))
                    {
                        throw new ArgumentException("The red and blue luminance weight values must sum to a floating-point value between 0 and 1.");
                    }
                    builder.Append("(");
                    builder.Append(bandCubeName);
                    builder.Append("[0,0,1] * ");
                    builder.Append(_LuminanceWeightRed.ToString());
                    builder.Append(") + (");
                    builder.Append(bandCubeName);
                    builder.Append("[0,0,2] * ");
                    builder.Append(weightGreen.ToString());
                    builder.Append(") + (");
                    builder.Append(bandCubeName);
                    builder.Append("[0,0,3] * ");
                    builder.Append(_LuminanceWeightBlue.ToString());
                    builder.Append(")");
                    //builder.Append(")))");
                    //builder.Append(Environment.NewLine);
                    break;
                default:
                    throw new Exception("Error: an unhandled BandCombineMode value was specified (" + _BandCombineMode.ToString() + ").");
                    break;
            }
            builder.Append("))");

            //builder.Append(GetNormalizationCodePost(OutputNormalization));

            builder.Append(Environment.NewLine);

            //builder.Append(OutputTransformations.GetWrappedLine(scriptTargetObjectName));



            // clear the temporary objects
            //for (int bandNum = 0; bandNum < _SpectralBands.Length; bandNum++)
            //{
            //    builder.Append(GreyscaleChannel.GetDeallocationScriptCode(_TemporaryObjectNamePrefix + String.Format("{0:000}", bandNum)));
            //    //builder.Append(_TemporaryObjectNamePrefix);
            //    //builder.Append(String.Format("{0:000}", bandNum));
            //    //builder.Append(_DeallocateObjectScriptCode);
            //    //builder.Append(Environment.NewLine);
            //}
            //if (generateBandCube)
            //{
            //    builder.Append(GetBandObjectDeallocationCode());
            //}

            if (mustGenerateMomentCube)
            {
                builder.Append(GetDeallocationScriptCode(scriptMomentCubeName));
            }

            builder.Append(GetChannelFileWriteCode(scriptTargetObjectName, regenerateExistingFiles));

            // de-allocate objects
            builder.Append(GreyscaleChannel.GetDeallocationScriptCode(scriptTargetObjectName));
            
            return builder.ToString();
        }

        public string GetBandCubeScriptCode(string scriptTargetObjectName)
        {
            StringBuilder builder = new StringBuilder();

            // read in the files
            for (int bandNum = 0; bandNum < _SpectralBands.Length; bandNum++)
            {
                string tempBandName = _TemporaryObjectNamePrefix + String.Format("cube_{0:000}", bandNum);
                builder.Append(tempBandName);
                builder.Append(" = ");

                StringBuilder b2 = new StringBuilder();

                //builder.Append(GetNormalizationCodePre(_SpectralBands[bandNum].InputNormalization));
                b2.Append(_CastOnLoad);
                b2.Append("(");
                b2.Append(_DataOrganizationMode);
                b2.Append("(read(\"");
                b2.Append(_SpectralBands[bandNum].GetPrimaryImageData().FileName);
                b2.Append("\")))");

                //builder.Append(GetNormalizationCodePost(_SpectralBands[bandNum].InputNormalization));

                //builder.Append(WrapScriptObjectInNormalizationCode(b2.ToString(), _SpectralBands[bandNum].InputNormalization));
                builder.Append(_SpectralBands[bandNum].InputTransformations.WrapScriptObjectInTransformationCode(b2.ToString()));

                builder.Append(Environment.NewLine);
                builder.Append(tempBandName);
                builder.Append(" = ");
                builder.Append(tempBandName);
                builder.Append(_SpectralBands[bandNum].GetPrimaryImageData().Plane);
                builder.Append(Environment.NewLine);
            }

            // concatenate them together along the z-axis
            // create the concatenation object and assign the first band to it
            builder.Append(scriptTargetObjectName);
            builder.Append(" = ");
            builder.Append(_TemporaryObjectNamePrefix);
            builder.Append("cube_000");
            builder.Append(Environment.NewLine);
            // concatenate in the other band(s)
            for (int bandNum = 1; bandNum < _SpectralBands.Length; bandNum++)
            {
                builder.Append(scriptTargetObjectName);
                builder.Append(" = cat(");
                builder.Append(scriptTargetObjectName);
                builder.Append(", ");
                builder.Append(_TemporaryObjectNamePrefix);
                builder.Append(String.Format("cube_{0:000}", bandNum));
                builder.Append(", axis = z)");
                builder.Append(Environment.NewLine);
            }

            // remove temporary objects
            builder.Append(GetBandObjectDeallocationCode());

            return builder.ToString();
        }

        public string GetMomentCubeScriptCode(string scriptTargetObjectName, string scriptSourceCubeObjectName, BandCombineMethod bandCombineMethods, int windowSize,
            TransformationProfileCollection cubeInputTransformations, TransformationProfileCollection cubeOutputTransformations)
        {
            // these variables are strings because they're written to the script as such;
            string calcMedian = "0";
            if (GreyscaleChannelSet.BandCombineMethodEnabled(bandCombineMethods, BandCombineMethod.Median))
            {
                calcMedian = "1";
            }
            string calcRange = "0";
            if (GreyscaleChannelSet.BandCombineMethodEnabled(bandCombineMethods, BandCombineMethod.Range))
            {
                calcRange = "1";
            }
            string calcFFAMin = "0";
            if (GreyscaleChannelSet.BandCombineMethodEnabled(bandCombineMethods, BandCombineMethod.FarthestFromAverageMin))
            {
                calcFFAMin = "1";
            }
            string calcFFAMax = "0";
            if (GreyscaleChannelSet.BandCombineMethodEnabled(bandCombineMethods, BandCombineMethod.FarthestFromAverageMax))
            {
                calcFFAMax = "1";
            }
            string calcFFMMin = "0";
            if (GreyscaleChannelSet.BandCombineMethodEnabled(bandCombineMethods, BandCombineMethod.FarthestFromMedianMin))
            {
                calcFFMMin = "1";
            }
            string calcFFMMax = "0";
            if (GreyscaleChannelSet.BandCombineMethodEnabled(bandCombineMethods, BandCombineMethod.FarthestFromMedianMax))
            {
                calcFFMMax = "1";
            }
            string calcCTAMin = "0";
            if (GreyscaleChannelSet.BandCombineMethodEnabled(bandCombineMethods, BandCombineMethod.ClosestToAverageMin))
            {
                calcCTAMin = "1";
            }
            string calcCTAMax = "0";
            if (GreyscaleChannelSet.BandCombineMethodEnabled(bandCombineMethods, BandCombineMethod.ClosestToAverageMax))
            {
                calcCTAMax = "1";
            }

            StringBuilder builder = new StringBuilder();

            if (cubeInputTransformations != null)
            {
                builder.Append(cubeInputTransformations.GetWrappedLine(scriptSourceCubeObjectName));
            }

            builder.Append(scriptTargetObjectName);
            builder.Append(" = momentcubeext(");
            builder.Append(scriptSourceCubeObjectName);
            builder.Append(", windowSize = ");
            builder.Append(windowSize.ToString());
            builder.Append(", calcMedian = ");
            builder.Append(calcMedian);
            builder.Append(", calcRange = ");
            builder.Append(calcRange);
            builder.Append(", calcFFAMin = ");
            builder.Append(calcFFAMin);
            builder.Append(", calcFFAMax = ");
            builder.Append(calcFFAMax);
            builder.Append(", calcFFMMin = ");
            builder.Append(calcFFMMin);
            builder.Append(", calcFFMMax = ");
            builder.Append(calcFFMMax);
            builder.Append(", calcCTAMin = ");
            builder.Append(calcCTAMin);
            builder.Append(", calcCTAMax = ");
            builder.Append(calcCTAMax);
            builder.Append(")");
            builder.Append(Environment.NewLine);

            if (cubeOutputTransformations != null)
            {
                builder.Append(cubeOutputTransformations.GetWrappedLine(scriptTargetObjectName));
            }

            return builder.ToString();
        }

        public string GetBandObjectDeallocationCode()
        {
            StringBuilder builder = new StringBuilder();
            // clear the temporary objects
            for (int bandNum = 0; bandNum < _SpectralBands.Length; bandNum++)
            {
                builder.Append(GreyscaleChannel.GetDeallocationScriptCode(_TemporaryObjectNamePrefix + String.Format("cube_{0:000}", bandNum)));
                //builder.Append(_TemporaryObjectNamePrefix);
                //builder.Append(String.Format("{0:000}", bandNum));
                //builder.Append(_DeallocateObjectScriptCode);
                //builder.Append(Environment.NewLine);
            }
            return builder.ToString();
        }

        protected string GetChannelFileWriteCode(string scriptTargetObjectName, bool regenerateExistingFiles)
        {
            StringBuilder builder = new StringBuilder();

            // normalize the image if that option is enabled
            //builder.Append(WrapScriptObjectInNormalizationCode(scriptTargetObjectName, OutputNormalization));
            //builder.Append(GetFullNormalizationWrappedLine(scriptTargetObjectName, OutputNormalization));
            //builder.Append(OutputTransformations.GetWrappedLine(scriptTargetObjectName));

            builder.Append(GetImageFileWriteCode(scriptTargetObjectName, OutputFormats, GetFilePathWithoutExtension(), regenerateExistingFiles));

            return builder.ToString();
        }

        public string GetFilePathWithoutExtension()
        {
            StringBuilder builder = new StringBuilder(OutputDirectory);
            builder.Append(_NamePrefix);
            builder.Append(ImageName);
            return builder.ToString();
        }

        public string GetFilePathWithPrimaryExtension()
        {
            if (OutputFormats.Length == 0)
            {
                throw new Exception("There are no output formats to obtain a file extension from.");
            }
            StringBuilder builder = new StringBuilder(OutputDirectory);
            builder.Append(_NamePrefix);
            builder.Append(ImageName);
            builder.Append(OutputParameterSet.ImageFormatToFileExtension(OutputFormats[0].FileFormat));
            return builder.ToString();
        }

        public bool ContainsBand(string bandName)
        {
            foreach (SpectralBand band in _SpectralBands)
            {
                if (band.BandName == bandName)
                {
                    return true;
                }
            }

            return false;
        }

        public SpectralBand GetBandByName(string bandName)
        {
            foreach (SpectralBand band in _SpectralBands)
            {
                if (band.BandName == bandName)
                {
                    return band;
                }
            }

            throw new ArgumentException("This channel does not contain a band named '" + bandName + "'.");
        }

        public bool ContainsChannel(string channelName)
        {
            foreach (GreyscaleChannel channel in ReferencedChannels)
            {
                if (channelName == channel.ImageName)
                {
                    return true;
                }
            }

            return false;
        }

        public GreyscaleChannel GetChannelByName(string channelName)
        {
            foreach (GreyscaleChannel channel in ReferencedChannels)
            {
                if (channelName == channel.ImageName)
                {
                    return channel;
                }
            }

            throw new ArgumentException("This channel does not contain a reference to a channel named '" + channelName + "'.");
        }

        public bool ContainsSpecialSpectralBand(SpecialBandName specialBand)
        {
            foreach (SpectralBand band in SpectralBands)
            {
                if (band.SpecialBandNames == specialBand)
                {
                    return true;
                }
            }
            return false;
        }

        public SpectralBand GetSpecialSpectralBand(SpecialBandName specialBand)
        {
            foreach (SpectralBand band in SpectralBands)
            {
                if (band.SpecialBandNames == specialBand)
                {
                    return band;
                }
            }
            throw new ArgumentException("The input configuration does not contain the special band '" + specialBand.ToString() + "'.");
        }

        public bool PrerequisitesExist()
        {
            try
            {
                GetCustomCodePrerequisites(false);
            }
            catch (Exception ex)
            {
                return false;
            }
            return true;
        }

        public void GetCustomCodePrerequisites()
        {
            GetCustomCodePrerequisites(true);
        }

        public void GetCustomCodePrerequisites(bool setProperties)
        {
            CustomChannelReference[] refs = GetPrerequisiteReferences();

            foreach (CustomChannelReference r in refs)
            {
                switch (r.ReferenceType)
                {
                    case CustomChannelReferenceType.BandByName:
                        if (ParentCube.ContainsBand(r.ReferenceName))
                        {
                            if (setProperties)
                            {
                                AddSpectralBand(ParentCube.GetSpectralBand(r.ReferenceName));
                                AddReferencedChannel(ParentCube.GetGreyscaleChannel(r.ReferenceName));
                            }
                        }
                        else
                        {
                            throw new ArgumentException("Could not find a spectral band named '" + r.ReferenceName + "' in the image cube.");
                        }
                        break;
                    case CustomChannelReferenceType.ChannelByName:
                        if (ParentCube.ContainsGreyscaleChannel(r.ReferenceName))
                        {
                            if (setProperties)
                            {
                                AddReferencedChannel(ParentCube.GetGreyscaleChannel(r.ReferenceName));
                            }
                        }
                        else
                        {
                            throw new ArgumentException("Could not find a channel named '" + r.ReferenceName + "' in the image cube.");
                        }
                        break;
                    case CustomChannelReferenceType.SpecialBand:
                        SpecialBandName sbName = SpectralBand.SpecialBandNameFromString(r.ReferenceName);
                        if (ParentCube.InputConfig.ContainsSpecialSpectralBand(sbName))
                        {
                            if (setProperties)
                            {
                                SpectralBand sb = ParentCube.InputConfig.GetSpecialSpectralBand(sbName);
                                AddSpectralBand(sb);
                                GreyscaleChannel newChan = ParentCube.GetGreyscaleChannel(sb.BandName);
                                AddReferencedChannel(newChan);
                                // so that the actual creation code doesn't need to know about the cube
                                string oldString = r.ScriptTextToReplace;
                                string newString = oldString.Replace(r.ReferenceName, newChan.ImageName);
                                newString = newString.Replace("SpecialBand:", "ChannelByName:");
                                _CustomBandCombinationScriptCode = _CustomBandCombinationScriptCode.Replace(oldString, newString);
                                //_CustomBandCombinationScriptCode = _CustomBandCombinationScriptCode.Replace("SpecialBand", "ChannelByName");
                            }
                        }
                        else
                        {
                            throw new ArgumentException("Could not find the special spectral band '" + r.ReferenceName + "' in the image cube.");
                        }
                        break;
                }
            }
        }

        protected CustomChannelReference[] GetPrerequisiteReferences()
        {
            ArrayList prereqs = new ArrayList();

            Regex rex = new Regex("%[^%]{1,1024}%");
            foreach (Match rexMatch in rex.Matches(CustomBandCombinationScriptCode))
            {
                CustomChannelReference newRef = new CustomChannelReference();
                newRef.SetPropertiesFromReferenceString(rexMatch.Value);
                prereqs.Add(newRef);
            }

            CustomChannelReference proto = new CustomChannelReference();
            return (CustomChannelReference[])prereqs.ToArray(proto.GetType());
        }

        //public string[] GetCustomCodePrerequisiteBandNames(ImageCube cube)
        //{
        //    ArrayList bandNames = new ArrayList();


        //    return (string[])bandNames.ToArray("proto".GetType());
        //}

        //public string[] GetCustomCodePrerequisiteChannelNames()
        //{
        //    if (CustomBandCombinationScriptCode == "")
        //    {
        //        return new string[0];
        //    }
        //    ArrayList channelNames = new ArrayList();


        //    return (string[])channelNames.ToArray("proto".GetType());
        //}

        public void AddReferencedChannel(GreyscaleChannel newChannel)
        {
            if ((ReferencedChannels == null) || (ReferencedChannels.Length == 0))
            {
                ReferencedChannels = new GreyscaleChannel[1];
            }
            else
            {
                Array.Resize(ref _ReferencedChannels, _ReferencedChannels.Length + 1);
            }
            ReferencedChannels[ReferencedChannels.Length - 1] = newChannel;
            foreach (SpectralBand newBand in newChannel.SpectralBands)
            {
                if (!ContainsBand(newBand))
                {
                    AddSpectralBand(newBand);
                }
            }
        }

        public void AddSpectralBand(SpectralBand newBand)
        {
            if ((SpectralBands == null) || (SpectralBands.Length == 0))
            {
                SpectralBands = new SpectralBand[1];
            }
            else
            {
                Array.Resize(ref _SpectralBands, _SpectralBands.Length + 1);
            }
            SpectralBands[SpectralBands.Length - 1] = newBand;
        }

        public bool ContainsBand(SpectralBand band)
        {
            return ContainsBand(band.BandName);
        }

        public bool ContainsAnyBand(SpectralBand[] bands)
        {
            foreach (SpectralBand band in bands)
            {
                if (ContainsBand(band))
                {
                    return true;
                }
            }
            return false;
        }

        public int GetMinimumBandNumber()
        {
            ArrayList bandNumbers = GetBandNumbers();
            return (int)bandNumbers[0];
        }

        public int GetMaximumBandNumber()
        {
            ArrayList bandNumbers = GetBandNumbers();
            return (int)bandNumbers[bandNumbers.Count - 1];
        }

        protected ArrayList GetBandNumbers()
        {
            ArrayList bandNumbers = new ArrayList(_SpectralBands.Length);
            foreach (SpectralBand band in _SpectralBands)
            {
                bandNumbers.Add(band.BandNumber);
            }
            bandNumbers.Sort();
            return bandNumbers;
        }



        //public void AddTintingColour(TintingColour newColour)
        //{
        //    if (_TintingColours.Contains(newColour.TintingColourName))
        //    {
        //        throw new ArgumentException("The channel '" + ImageName + "' already has a tinting colour named '" + newColour.TintingColourName + "' associated with it.");
        //    }
        //    _TintingColours.Add(newColour.TintingColourName, newColour);
        //}

        //public TintingColour GetTintingColour(string colourName)
        //{
        //    if (!_TintingColours.Contains(colourName))
        //    {
        //        throw new ArgumentException("The channel '" + ImageName + "' does not have a tinting colour named '" + colourName + "' associated with it.");
        //    }
        //    return (TintingColour)_TintingColours[colourName];
        //}

        //public bool ContainsColour(string colourName)
        //{
        //    return _TintingColours.Contains(colourName);
        //}

        //public string[] GetColourNames()
        //{
        //    TintingColour[] colourList = GetSortedColours();
        //    string[] colourNames = new string[colourList.Length];
        //    for (int i = 0; i < colourList.Length; i++)
        //    {
        //        TintingColour currentColour = (TintingColour)colourList[i];
        //        colourNames[i] = currentColour.TintingColourName;
        //    }
        //    return colourNames;
        //}

        //public TintingColour[] GetSortedColours()
        //{
        //    if (_TintingColours.Count == 0)
        //    {
        //        return new TintingColour[0] { };
        //    }
        //    ArrayList colourList = new ArrayList();
        //    foreach (string key in _TintingColours.Keys)
        //    {
        //        colourList.Add(GetTintingColour(key));
        //    }
        //    colourList.Sort();
        //    TintingColour b = new TintingColour("proto", "proto", 0.0f, 0.0f, 0.0f);
        //    return (TintingColour[])colourList.ToArray(b.GetType());
        //}

        public int CompareTo(object obj)
        {
            if (obj is GreyscaleChannel)
            {
                GreyscaleChannel b2 = (GreyscaleChannel)obj;
                return ImageName.CompareTo(b2.ImageName);
            }
            else
            {
                return ImageName.CompareTo(obj.ToString());
            }
        }

        //public static string GetDeallocationScriptCode(string scriptTargetObjectName)
        //{
        //    return scriptTargetObjectName + _DeallocateObjectScriptCode + Environment.NewLine;
        //}

        public override int GetMaximumFileCount()
        {
            //return OutputFormats.Length * (1 + _TintingColours.Count);
            return OutputFormats.Length;
        }

        public bool HasMutuallyExclusiveBands()
        {
            foreach (SpectralBand band in SpectralBands)
            {
                foreach (SpectralBand otherBand in SpectralBands)
                {
                    if ((band.BandName != otherBand.BandName) && ((band.IsMutuallyExclusiveWithBand(otherBand.BandName)) || (otherBand.IsMutuallyExclusiveWithBand(band.BandName))))
                    {
                        return true;
                    }
                }
            }
            return false;
        }

        public bool HasMutuallyExclusiveBands(GreyscaleChannel otherChannel)
        {
            if (HasMutuallyExclusiveBands())
            {
                return true;
            }
            return _HasMutuallyExclusiveBands(otherChannel);
        }

        protected bool _HasMutuallyExclusiveBands(GreyscaleChannel otherChannel)
        {
            foreach (SpectralBand band in SpectralBands)
            {
                foreach (SpectralBand ocBand in otherChannel.SpectralBands)
                {
                    if ((band.IsMutuallyExclusiveWithBand(ocBand.BandName)) || (ocBand.IsMutuallyExclusiveWithBand(band.BandName)))
                    {
                        return true;
                    }
                }
            }
            return false;
        }

        public bool HasMutuallyExclusiveBands(GreyscaleChannel[] otherChannels)
        {
            if (HasMutuallyExclusiveBands())
            {
                return true;
            }
            foreach (GreyscaleChannel otherChannel in otherChannels)
            {
                if (_HasMutuallyExclusiveBands(otherChannel))
                {
                    return true;
                }
            }
            return false;
        }

        public bool UseAsLumaBand()
        {
            foreach (SpectralBand band in SpectralBands)
            {
                if (!band.UseAsLumaChannel)
                {
                    return false;
                }
            }
            return true;
        }

        //public bool UseAsThreeChannelChannel()
        //{
        //    foreach (SpectralBand band in SpectralBands)
        //    {
        //        if (!band.UseAsThreeChannelChannel)
        //        {
        //            return false;
        //        }
        //    }
        //    return true;
        //}

        public static string GetBandCombineMethodNameString(BandCombineMethod method)
        {
            switch (method)
            {
                case BandCombineMethod.Average:
                    return "Avg";
                    break;
                case BandCombineMethod.Minimum:
                    return "Min";
                    break;
                case BandCombineMethod.Maximum:
                    return "Max";
                    break;
                case BandCombineMethod.AverageDeviation:
                    return "ADev";
                    break;
                case BandCombineMethod.StandardDeviation:
                    return "SDev";
                    break;
                case BandCombineMethod.Variance:
                    return "Var";
                    break;
                case BandCombineMethod.Skewness:
                    return "Skew";
                    break;
                case BandCombineMethod.Kurtosis:
                    return "Kurt";
                    break;
                case BandCombineMethod.Median:
                    return "Med";
                    break;
                case BandCombineMethod.Range:
                    return "Range";
                    break;
                case BandCombineMethod.FarthestFromAverageMin:
                    return "FFAMin";
                    break;
                case BandCombineMethod.FarthestFromAverageMax:
                    return "FFAMax";
                    break;
                case BandCombineMethod.FarthestFromMedianMin:
                    return "FFMMin";
                    break;
                case BandCombineMethod.FarthestFromMedianMax:
                    return "FFMMax";
                    break;
                case BandCombineMethod.ClosestToAverageMin:
                    return "CTAMin";
                    break;
                case BandCombineMethod.ClosestToAverageMax:
                    return "CTAMax";
                    break;
                case BandCombineMethod.Luminance:
                    return "Luma";
                    break;
            }
            throw new ArgumentException("Could not convert the BandCombineMethod '" + method.ToString() + "' to a string - its type is not recognized by this function.");
        }

        public static BandCombineMethod GetBandCombineMethodFromString(string methodName)
        {
            string mName = methodName.Trim().ToLower();
            switch (mName)
            {
                case "avg":
                    return BandCombineMethod.Average;
                    break;
                case "average":
                    return BandCombineMethod.Average;
                    break;
                case "min":
                    return BandCombineMethod.Minimum;
                    break;
                case "minimum":
                    return BandCombineMethod.Minimum;
                    break;
                case "max":
                    return BandCombineMethod.Maximum;
                    break;
                case "maximum":
                    return BandCombineMethod.Maximum;
                    break;
                case "adev":
                    return BandCombineMethod.AverageDeviation;
                    break;
                case "avgdev":
                    return BandCombineMethod.AverageDeviation;
                    break;
                case "averagedeviation":
                    return BandCombineMethod.AverageDeviation;
                    break;
                case "sdev":
                    return BandCombineMethod.StandardDeviation;
                    break;
                case "stddev":
                    return BandCombineMethod.StandardDeviation;
                    break;
                case "standarddeviation":
                    return BandCombineMethod.StandardDeviation;
                    break;
                case "var":
                    return BandCombineMethod.Variance;
                    break;
                case "variance":
                    return BandCombineMethod.Variance;
                    break;
                case "skew":
                    return BandCombineMethod.Skewness;
                    break;
                case "skewness":
                    return BandCombineMethod.Skewness;
                    break;
                case "kurt":
                    return BandCombineMethod.Kurtosis;
                    break;
                case "kurtosis":
                    return BandCombineMethod.Kurtosis;
                    break;
                case "med":
                    return BandCombineMethod.Median;
                    break;
                case "median":
                    return BandCombineMethod.Median;
                    break;
                case "range":
                    return BandCombineMethod.Range;
                    break;
                case "ffamin":
                    return BandCombineMethod.FarthestFromAverageMin;
                    break;
                case "ffamax":
                    return BandCombineMethod.FarthestFromAverageMax;
                    break;
                case "ffmmin":
                    return BandCombineMethod.FarthestFromMedianMin;
                    break;
                case "ffmmax":
                    return BandCombineMethod.FarthestFromMedianMax;
                    break;
                case "ctamin":
                    return BandCombineMethod.ClosestToAverageMin;
                    break;
                case "ctamax":
                    return BandCombineMethod.ClosestToAverageMax;
                    break;
                case "ratio":
                    return BandCombineMethod.Ratio;
                    break;
                case "difference":
                    return BandCombineMethod.Difference;
                    break;
                case "normalizeddifference":
                    return BandCombineMethod.NormalizedDifference;
                    break;
                case "normalizeddifferencecenteredonzeropointfive":
                    return BandCombineMethod.NormalizedDifferenceCenteredOnZeroPointFive;
                    break;
                case "luma":
                    return BandCombineMethod.Luminance;
                    break;
                case "luminance":
                    return BandCombineMethod.Luminance;
                    break;
                case "custom":
                    return BandCombineMethod.Custom;
                    break;
            }
            throw new ArgumentException("Could not convert the string '" + methodName + "' to a recognized BandCombineMethod.");
        }

    }
}
