// BenLincoln.DaVinci - A shared library used by The Mirror's Surface Breaks
// Copyright 2010-2011 Ben Lincoln
// http://www.beneaththewaves.net/
//

// This file is part of BenLincoln.DaVinci.

// BenLincoln.DaVinci is free software: you can redistribute it and/or modify
// it under the terms of version 3 of the GNU General Public License as published by
// the Free Software Foundation.

// BenLincoln.DaVinci is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with BenLincoln.DaVinci (in the file LICENSE.txt).  
// If not, see <http://www.gnu.org/licenses/>.

using System;
using System.Collections.Generic;
using System.IO;
using System.Text;
using System.Threading;

using BenLincoln.Utility;

namespace BenLincoln.DaVinci
{
    public class TMSBCommandLineEngineMonitor : TMSBEngineMonitor, IDisposable
    {
        protected bool _LogToConsole;

        protected bool _LogToFile;
        protected bool _OverwriteExistingFile;
        protected string _LogFilePath;

        protected bool _LogToStream;
        protected Stream _LogStream;

        protected int _FileIOErrorCount;
        protected int _StreamIOErrorCount;

        protected const int _SleepLengthOutputErrorRetry = 5000;
        protected const int _SleepLengthWaitingForWriteLock = 10;
        protected const int _MaxOutputErrorRetries = 4;
        protected const int _MaxOutputErrorCount = 16;

        #region Properties

        public bool LogToConsole
        {
            get
            {
                return _LogToConsole;
            }
            set
            {
                _LogToConsole = value;
            }
        }

        public bool LogToFile
        {
            get
            {
                return _LogToFile;
            }
        }

        public bool OverwriteExistingFile
        {
            get
            {
                return _OverwriteExistingFile;
            }
        }

        public string LogFilePath
        {
            get
            {
                return _LogFilePath;
            }
        }

        public bool LogToStream
        {
            get
            {
                return _LogToStream;
            }
        }

        public Stream LogStream
        {
            get
            {
                return _LogStream;
            }
        }

        #endregion

        public TMSBCommandLineEngineMonitor(TMSBEngineController controller, OutputLevel outputLevel)
            : base(controller, outputLevel)
        {
            // locked so that all event handlers are registered before logging begins
            lock (GlobalLock.LockObject)
            {
                _LogToConsole = true;
                _LogToFile = false;
                _OverwriteExistingFile = false;
                _LogFilePath = "";
                _LogToStream = false;

                _FileIOErrorCount = 0;
                _StreamIOErrorCount = 0;

                if (!controller.EngineControllerConfiguration.WriteOutputToConsole)
                {
                    _LogToConsole = false;
                }

                if ((controller.EngineControllerConfiguration.WriteOutputToLogFile) &&
                    (controller.EngineControllerConfiguration.LogFilePath != ""))
                {
                    bool overwriteExisting = true;
                    if (controller.EngineControllerConfiguration.LogFileMode == LogMode.Append)
                    {
                        overwriteExisting = false;
                    }
                    EnableFileLogging(controller.EngineControllerConfiguration.LogFilePath,
                        overwriteExisting);
                }

                _MetaTracker.TotalElementCount = controller.BaseDirectories.Length;
                if (outputLevel > OutputLevel.Verbose)
                {
                    controller.EngineControllerDebugMessage += OnEngineControllerDebugMessageGenerated;
                    controller.ScriptExecutionDebugMessageHandler += OnScriptDebugMessageGenerated;
                    controller.WaitingForEngineThread += OnGeneralThreadManagementEvent;
                    controller.ObtainedEngineThread += OnGeneralThreadManagementEvent;
                    controller.WaitingForProcessingThreadHandler += OnGeneralThreadManagementEvent;
                    controller.ObtainedProcessingThreadHandler += OnGeneralThreadManagementEvent;
                }
                if (outputLevel > OutputLevel.Standard)
                {
                    controller.EngineControllerInformationalMessage += OnGeneralThreadManagementEvent;
                    controller.ScriptExecutionInformationalMessageHandler += OnScriptInformationalMessageGenerated;
                }
                if (outputLevel > OutputLevel.Minimal)
                {
                    controller.CubeScriptGeneratedHandler += OnCubeScriptGenerated;
                    controller.EngineImageStartedHandler += OnImageStarted;
                    controller.EngineImageFinishedHandler += OnImageFinished;
                }
                if (outputLevel > OutputLevel.None)
                {
                    controller.DirectoryProcessingStartedHandler += OnDirectoryProcessingStarted;
                    controller.ScriptExecutionCompleteHandler += OnScriptExecutionComplete;
                    controller.DirectoryProcessingFinishedHandler += OnDirectoryProcessingFinished;
                    controller.EngineControllerExecutionStarted += OnEngineControllerExecutionStarted;
                    controller.EngineControllerExecutionComplete += OnEngineControllerExecutionComplete;
                }
                controller.ScriptExecutionAbortedHandler += OnScriptExecutionAborted;
                controller.ScriptExecutionErrorHandler += OnScriptExecutionError;
                controller.UnhandledExceptionHandler += OnUnhandledExceptionThrown;
            }
        }



        protected override void Dispose(bool disposing)
        {
            // If you need thread safety, use a lock around these 
            // operations, as well as in your methods that use the resource.
            if (!_Disposed)
            {
                //if (disposing)
                //{
                    if (_OutputLevel > OutputLevel.Verbose)
                    {
                        _Controller.EngineControllerDebugMessage -= OnEngineControllerDebugMessageGenerated;
                        _Controller.ScriptExecutionDebugMessageHandler -= OnScriptDebugMessageGenerated;
                        _Controller.WaitingForEngineThread -= OnGeneralThreadManagementEvent;
                        _Controller.ObtainedEngineThread -= OnGeneralThreadManagementEvent;
                        _Controller.WaitingForProcessingThreadHandler -= OnGeneralThreadManagementEvent;
                        _Controller.ObtainedProcessingThreadHandler -= OnGeneralThreadManagementEvent;
                    }
                    if (_OutputLevel > OutputLevel.Standard)
                    {
                        _Controller.EngineControllerInformationalMessage -= OnGeneralThreadManagementEvent;
                        _Controller.ScriptExecutionInformationalMessageHandler -= OnScriptInformationalMessageGenerated;
                    }
                    if (_OutputLevel > OutputLevel.Minimal)
                    {
                        _Controller.CubeScriptGeneratedHandler -= OnCubeScriptGenerated;
                        _Controller.EngineImageStartedHandler -= OnImageStarted;
                        _Controller.EngineImageFinishedHandler -= OnImageFinished;
                    }
                    if (_OutputLevel > OutputLevel.None)
                    {
                        _Controller.DirectoryProcessingStartedHandler -= OnDirectoryProcessingStarted;
                        _Controller.ScriptExecutionCompleteHandler -= OnScriptExecutionComplete;
                        _Controller.DirectoryProcessingFinishedHandler -= OnDirectoryProcessingFinished;
                        _Controller.EngineControllerExecutionStarted -= OnEngineControllerExecutionStarted;
                        _Controller.EngineControllerExecutionComplete -= OnEngineControllerExecutionComplete;
                    }
                    _Controller.ScriptExecutionAbortedHandler -= OnScriptExecutionAborted;
                    _Controller.ScriptExecutionErrorHandler -= OnScriptExecutionError;
                    _Controller.UnhandledExceptionHandler -= OnUnhandledExceptionThrown;
                //}

                //_Disposed = true;
            }
            base.Dispose(disposing);
        }

        public void EnableFileLogging(string logFilePath, bool overwriteExistingFile)
        {
            _LogToFile = true;
            _LogFilePath = logFilePath;
            _OverwriteExistingFile = overwriteExistingFile;
            _FileIOErrorCount = 0;
            if (overwriteExistingFile)
            {
                CreateLogFile(logFilePath);
            }
        }

        protected static void CreateLogFile(string logFilePath)
        {
            try
            {
                FileStream tStream = new FileStream(logFilePath, FileMode.Create, FileAccess.Write);
                StreamWriter tWriter = new StreamWriter(tStream);
                tWriter.Write("");
                tWriter.Close();
                tStream.Close();
            }
            catch (Exception ex)
            {
                ThrowFileIOException(logFilePath, "create the log file", ex);
            }
        }

        public void DisableFileLogging()
        {
            _LogToFile = false;
            _LogFilePath = "";
            _OverwriteExistingFile = false;
        }

        public void EnableStreamLogging(ref Stream logStream)
        {
            _LogToStream = true;
            _LogStream = logStream;
            _StreamIOErrorCount = 0;
        }

        public void DisableStreamLogging()
        {
            _LogToStream = false;
            _LogStream = null;
        }

        protected void WriteOutput(int engineNumber, int globalThreadNumber, string message)
        {
            bool disableFileLogging = false;
            bool disableStreamLogging = false;

            WriteOutput(_Controller, engineNumber, globalThreadNumber, message,
                _LogToConsole, _LogToFile, _LogToStream, _LogFilePath, _LogStream,
                out disableFileLogging, out disableStreamLogging,
                ref _FileIOErrorCount, ref _StreamIOErrorCount);

            if (disableFileLogging)
            {
                DisableFileLogging();
            }

            if (disableStreamLogging)
            {
                DisableStreamLogging();
            }
        }

        protected static void WriteOutput(TMSBEngineController ec, int engineNumber, int globalThreadNumber, string message,
            bool logToConsole, bool logToFile, bool logToStream, string logFilePath, Stream logStream,
            out bool disableFileLogging, out bool disableStreamLogging, 
            ref int fileIOErrorCount, ref int streamIOErrorCount)
        {
            // locked to prevent multiple lines being written to the console or log file simultaneously
            lock(GlobalLock.LockObject)
            {
                disableFileLogging = false;
                disableStreamLogging = false;
                StringBuilder builder = new StringBuilder();
                builder.Append(GetCurrentDateTimeString());
                builder.Append("\t");
                if ((engineNumber != -1) && (ec.EngineControllerConfiguration.EngineThreadCount > 1))
                {
                    builder.Append(String.Format("Engine {0:00}\t", engineNumber));
                }
                if ((globalThreadNumber != -1) && (ec.EngineControllerConfiguration.GlobalThreadCount > 1))
                {
                    builder.Append(String.Format("Thread {0:00}\t", globalThreadNumber));
                }
                builder.Append(message);

                string outputLine = builder.ToString();

                if (logToConsole)
                {
                    Console.WriteLine(outputLine);
                }
                int retryCount = 0;
                if ((logToFile) && (logFilePath != ""))
                {
                    bool writtenToFile = false;
                    retryCount = 0;
                    while (!writtenToFile)
                    {
                        try
                        {
                            if (!File.Exists(logFilePath))
                            {
                                CreateLogFile(logFilePath);
                            }
                            FileStream oStream = new FileStream(logFilePath, FileMode.Append, FileAccess.Write);
                            StreamWriter oWriter = new StreamWriter(oStream);
                            oWriter.WriteLine(outputLine);
                            oWriter.Close();
                            oStream.Close();
                            try
                            {
                                oWriter.Dispose();
                                oStream.Dispose();
                            }
                            catch (Exception ex)
                            {
                                // do nothing
                            }
                            writtenToFile = true;
                        }
                        catch (Exception ex)
                        {
                            retryCount++;
                            if (retryCount > _MaxOutputErrorRetries)
                            {
                                fileIOErrorCount++;
                                if (fileIOErrorCount > _MaxOutputErrorCount)
                                {
                                    disableFileLogging = true;
                                    Console.WriteLine("File IO Error: Unable to write to the log file '" +
                                        logFilePath + "'. Retried " + fileIOErrorCount.ToString() + " messages " +
                                        _MaxOutputErrorRetries.ToString() + " times. Log file output has been disabled.");
                                    break;
                                }
                            }
                            Thread.Sleep(_SleepLengthOutputErrorRetry);
                        }
                    }
                }
                if ((logToStream) && (logStream != null))
                {
                    if (!logStream.CanWrite)
                    {
                        disableStreamLogging = true;
                        Console.WriteLine("Stream IO Error: Unable to write to the output stream. Stream output has been disabled.");
                    }
                    bool writtenToStream = false;
                    retryCount = 0;
                    while (!writtenToStream)
                    {
                        try
                        {
                            StreamWriter oWriter = new StreamWriter(logStream);
                            oWriter.WriteLine(outputLine);
                            writtenToStream = true;
                        }
                        catch (Exception ex)
                        {
                            retryCount++;
                            if (retryCount > _MaxOutputErrorRetries)
                            {
                                streamIOErrorCount++;
                                if (streamIOErrorCount > _MaxOutputErrorCount)
                                {
                                    disableStreamLogging = true;
                                    Console.WriteLine("Stream IO Error: Unable to write to the output stream ." +
                                        "Retried " + streamIOErrorCount.ToString() + " messages " +
                                        _MaxOutputErrorRetries.ToString() + " times. Stream output has been disabled.");
                                    break;
                                }
                            }
                            Thread.Sleep(_SleepLengthOutputErrorRetry);
                        }
                    }
                }
            }
        }

        public void OnGeneralThreadManagementEvent(object sender, EngineManagementEventArgs e)
        {
            WriteOutput(e.EngineNumber, e.GlobalThreadNumber, e.Message);
        }

        public void OnEngineControllerExecutionError(object sender, EngineManagementEventArgs e)
        {
            WriteOutput(e.EngineNumber, e.GlobalThreadNumber, e.Message);
        }

        public void OnEngineControllerDebugMessageGenerated(object sender, EngineManagementEventArgs e)
        {
            WriteOutput(e.EngineNumber, e.GlobalThreadNumber, "Debug: " + e.Message);
        }

        public void OnScriptDebugMessageGenerated(object sender, ScriptExecutionEventArgs s)
        {
            WriteOutput(s.EngineNumber, s.GlobalThreadNumber, "Debug: " + s.Message);
        }

        public void OnScriptInformationalMessageGenerated(object sender, ScriptExecutionEventArgs s)
        {
            WriteOutput(s.EngineNumber, s.GlobalThreadNumber, s.Message);
        }

        public void OnUnhandledExceptionThrown(object sender, UnhandledExceptionArgs u)
        {
            WriteOutput(u.EngineNumber, u.GlobalThreadNumber, "Unhandled exception: " + u.Exception.Message);
        }

        public void OnScriptExecutionError(object sender, ScriptExecutionErrorArgs e)
        {
            StringBuilder builder = new StringBuilder();
            builder.Append("Error: Script '");
            builder.Append(e.ScriptPath);
            builder.Append("' (execution directory: '");
            builder.Append(e.DirectoryPath);
            builder.Append("') generated ");
            builder.Append(String.Format("{0}", e.ErrorMessages.Length));
            if (e.ErrorMessages.Length == 1)
            {
                builder.Append(" error. ");
            }
            else
            {
                builder.Append(" errors. ");
            }
            if (_Controller.EngineControllerConfiguration.HaltOnErrors)
            {
                builder.Append("Execution of the script was halted due to the errors. Output for this directory will be incomplete.");
            }
            else
            {
                builder.Append("Execution of the script continued, despite the errors. Output for this directory is likely to be incomplete and/or corrupt.");
            }
            WriteOutput(e.EngineNumber, e.GlobalThreadNumber, builder.ToString());
            if (_Controller.EngineControllerConfiguration.OutputLevel == OutputLevel.Verbose)
            {
                WriteOutput(e.EngineNumber, e.GlobalThreadNumber, "Error messages returned: ");
                foreach (string errMsg in e.ErrorMessages)
                {
                    WriteOutput(e.EngineNumber, e.GlobalThreadNumber, errMsg);
                }
            }
        }

        public void OnDirectoryProcessingStarted(object sender, DirectoryProcessingArgs d)
        {
            lock (_MetaTracker)
            {
                _MetaTracker.CurrentElementNumber++;
            }
            WriteOutput(d.EngineNumber, d.GlobalThreadNumber, "Processing " + GetDirectoryInfoString(d));
        }

        public void OnDirectoryProcessingFinished(object sender, DirectoryProcessingArgs d)
        {
            WriteOutput(d.EngineNumber, d.GlobalThreadNumber, "Finished processing " + GetDirectoryInfoString(d));
            //lock (_MetaTracker)
            //{
                //_MetaTracker.AddElapsedTime(_Controller.Engines[d.EngineNumber].Engine.Tracker.GetTotalElapsedTime());
                _MetaTracker.AddElapsedTime(d.ElapsedTime);
            //}
            TimeRemainingInformation tri = _MetaTracker.GetTimeRemainingInformation();
            WriteOutput(d.EngineNumber, d.GlobalThreadNumber, "Total elapsed time for this batch: " + tri.TotalElapsedTimeLongString);
            WriteOutput(d.EngineNumber, d.GlobalThreadNumber, "Average time per directory: " + tri.AverageTimePerElementLongString);
            WriteOutput(d.EngineNumber, d.GlobalThreadNumber, "Median time per directory: " + tri.MedianTimePerElementLongString);

            WritePerformanceStatistics(d.EngineNumber, d.GlobalThreadNumber, "While Processing This Directory",
                _Controller.Engines[d.EngineNumber].Engine.MemoryUseTracker, _Controller.Engines[d.EngineNumber].Engine.ProcessorTimeTracker);

            int remainingElements = _MetaTracker.TotalElementCount - 
                    _MetaTracker.CurrentElementNumber;

            if (DisplayEstimatedTimeRemaining(tri, remainingElements))
            {
                WriteOutput(d.EngineNumber, d.GlobalThreadNumber, "Estimated time remaining for all directories: " + tri.EstimatedTimeRemainingLongString);
            }
        }

        public void OnCubeScriptGenerated(object sender, CubeScriptGeneratedEventArgs c)
        {
            WriteOutput(c.EngineNumber, c.GlobalThreadNumber, "Script execution beginning on " + _Controller.Engines[c.EngineNumber].Engine.Tracker.StartTime.ToLongDateString() +
                " at " + _Controller.Engines[c.EngineNumber].Engine.Tracker.StartTime.ToLongTimeString());
        }

        public void OnScriptExecutionComplete(object sender, ScriptExecutionEventArgs s)
        {
            TimeRemainingInformation tri = _Controller.Engines[s.EngineNumber].Engine.Tracker.GetTimeRemainingInformation();

            WriteOutput(s.EngineNumber, s.GlobalThreadNumber, "Script execution ('" + s.Message +
                "') complete on " + _Controller.Engines[s.EngineNumber].Engine.Tracker.EndTime.ToLongDateString() +
                " at " + _Controller.Engines[s.EngineNumber].Engine.Tracker.EndTime.ToLongTimeString());
            WriteOutput(s.EngineNumber, s.GlobalThreadNumber, "Total execution time: " + tri.TotalElapsedTimeLongString);
        }

        public void OnScriptExecutionAborted(object sender, ScriptExecutionEventArgs s)
        {
            WriteOutput(s.EngineNumber, s.GlobalThreadNumber, "Script execution aborted on " + _Controller.Engines[s.EngineNumber].Engine.Tracker.EndTime.ToLongDateString() +
                " at " + _Controller.Engines[s.EngineNumber].Engine.Tracker.EndTime.ToLongTimeString() + " Reason: '" + s.Message + "'.");
        }

        public void OnImageStarted(object sender, ImageEventArgs a)
        {
            if (a.FirstImageNumber > 0)
            {
                if (_CurrentSection != a.SectionName)
                {
                    _CurrentSection = a.SectionName;
                    if (_OutputLevel == OutputLevel.Verbose)
                    {
                        WriteOutput(a.EngineNumber, a.GlobalThreadNumber, "Beginning section '" + _CurrentSection + "'");
                    }
                }
                if (_OutputLevel == OutputLevel.Verbose)
                {
                    WriteOutput(a.EngineNumber, a.GlobalThreadNumber, GetImageStatusLine(a));
                }
            }
        }

        public void OnImageFinished(object sender, ImageFinishedEventArgs a)
        {
            if (a.FirstImageNumber > 0)
            {
                WriteOutput(a.EngineNumber, a.GlobalThreadNumber, GetImageStatusLine(a));

                WriteOutput(a.EngineNumber, a.GlobalThreadNumber, "\tElapsed Time For This Image: " + TimeTracker.GetLongFormattedTimeSpanString(a.ElapsedTime));

                TimeRemainingInformation tri = _Controller.Engines[a.EngineNumber].Engine.Tracker.GetTimeRemainingInformation();

                if (_OutputLevel > OutputLevel.Standard)
                {
                    if (_Controller.Engines[a.EngineNumber].Engine.Tracker.Values.Length > 0)
                    {
                        WriteOutput(a.EngineNumber, a.GlobalThreadNumber, "Average Time Per Image: " + tri.AverageTimePerElementLongString);
                        WriteOutput(a.EngineNumber, a.GlobalThreadNumber, "Median Time Per Image: " + tri.MedianTimePerElementLongString);
                    }

                    WritePerformanceStatistics(a.EngineNumber, a.GlobalThreadNumber, "While Generating This Image",
                        a.MemoryUseTracker, a.ProcessorTimeTracker);

                    WriteOutput(a.EngineNumber, a.GlobalThreadNumber, "Current Elapsed Time: " + tri.TotalElapsedTimeLongString);
                }

                int remainingElements = _Controller.Engines[a.EngineNumber].Engine.Tracker.TotalElementCount -
                    _Controller.Engines[a.EngineNumber].Engine.Tracker.CurrentElementNumber;

                if (DisplayEstimatedTimeRemaining(tri, remainingElements))
                {
                    WriteOutput(a.EngineNumber, a.GlobalThreadNumber, "Estimated Remaining Time: " + tri.EstimatedTimeRemainingLongString);
                }
            }
        }

        protected string GetImageStatusLine(ImageEventArgs a)
        {
            StringBuilder builder = new StringBuilder();
            builder.Append("\t");
            int currentImageNumber = _Controller.Engines[a.EngineNumber].Engine.Tracker.CurrentElementNumber;
            if (a is ImageFinishedEventArgs)
            {
                builder.Append("Finished ");
                ImageFinishedEventArgs ifea = (ImageFinishedEventArgs)a;
                currentImageNumber = ifea.GlobalImageNumber;
                //if (ifea.TotalImages > 1)
                //{
                //    currentImageNumber += (ifea.TotalImages - 1);
                //}
            }
            else
            {
                builder.Append("Beginning ");
            }
            //builder.Append(String.Format("image {0}", a.ImageNumber));
            builder.Append(String.Format("image {0}", currentImageNumber));
            //if (a.TotalImages > 0)
            if (_Controller.Engines[a.EngineNumber].Engine.Tracker.TotalElementCount > 0)
            {
                //builder.Append(String.Format(" of {0}", a.TotalImages));
                builder.Append(String.Format(" of {0}", _Controller.Engines[a.EngineNumber].Engine.Tracker.TotalElementCount));
            }
            builder.Append(" - '");
            builder.Append(a.ImageName);
            builder.Append("'");
            return builder.ToString();
        }

        public void OnEngineControllerExecutionStarted(object sender, EngineManagementEventArgs e)
        {
            if (e.Message != "")
            {
                WriteOutput(-1, -1, e.Message);
            }
        }

        public void OnEngineControllerExecutionComplete(object sender, EngineManagementEventArgs e)
        {
            if (e.Message != "")
            {
                WriteOutput(-1, -1, e.Message);
            }
            else
            {
                WriteOutput(-1, -1, "All directories have been processed. Execution complete.");
            }
            //if (_Controller.EngineControllerConfiguration.OutputLevel > OutputLevel.Standard)
            //{
            //    if (_Controller.MemoryUseTracker.Values.Length > 0)
            //    {
            //        WriteOutput(e.EngineNumber, e.GlobalThreadNumber, "Maximum System Memory Used During Processing: " +
            //            _Controller.GetMemoryUsageString(_Controller.MemoryUseTracker.Maximum));
            //        WriteOutput(e.EngineNumber, e.GlobalThreadNumber, "Average System Memory Used During Processing: " +
            //            _Controller.GetMemoryUsageString(_Controller.MemoryUseTracker.GetAverageValue()));
            //        WriteOutput(e.EngineNumber, e.GlobalThreadNumber, "Median System Memory Used During Processing: " +
            //            _Controller.GetMemoryUsageString(_Controller.MemoryUseTracker.GetMedianValue()));
            //    }
            //}
            WritePerformanceStatistics(e.EngineNumber, e.GlobalThreadNumber, "During Processing",
                _Controller.MemoryUseTracker, _Controller.ProcessorTimeTracker);
        }

        public void WritePerformanceStatistics(int engineNumber, int globalThreadNumber, string activityDescription, 
            NumericValueTracker memoryUseTracker, NumericValueTracker processorTimeTracker)
        {
            if (!_Controller.EngineControllerConfiguration.CollectPerformanceData)
            {
                return;
            }
            if (_Controller.EngineControllerConfiguration.OutputLevel > OutputLevel.Standard)
            {
                if (memoryUseTracker.Values.Length > 0)
                {
                    WriteOutput(engineNumber, globalThreadNumber, "Maximum System Memory Used " + activityDescription + ": " +
                        _Controller.GetMemoryUsageString(memoryUseTracker.Maximum));
                    WriteOutput(engineNumber, globalThreadNumber, "Average System Memory Used " + activityDescription + ": " +
                        _Controller.GetMemoryUsageString(memoryUseTracker.GetAverageValue()));
                    WriteOutput(engineNumber, globalThreadNumber, "Median System Memory Used " + activityDescription + ": " +
                        _Controller.GetMemoryUsageString(memoryUseTracker.GetMedianValue()));
                }
                if (processorTimeTracker.Values.Length > 0)
                {
                    WriteOutput(engineNumber, globalThreadNumber, "Maximum Percent Processor Time Used " + activityDescription + ": " +
                        _Controller.GetPercentProcessorTimeString(processorTimeTracker.Maximum));
                    WriteOutput(engineNumber, globalThreadNumber, "Average Percent Processor Time Used " + activityDescription + ": " +
                        _Controller.GetPercentProcessorTimeString(processorTimeTracker.GetAverageValue()));
                    WriteOutput(engineNumber, globalThreadNumber, "Median Percent Processor Time Used " + activityDescription + ": " +
                        _Controller.GetPercentProcessorTimeString(processorTimeTracker.GetMedianValue()));
                }
            }
        }
    }
}
