// BenLincoln.DaVinci - A shared library used by The Mirror's Surface Breaks
// Copyright 2010-2011 Ben Lincoln
// http://www.beneaththewaves.net/
//

// This file is part of BenLincoln.DaVinci.

// BenLincoln.DaVinci is free software: you can redistribute it and/or modify
// it under the terms of version 3 of the GNU General Public License as published by
// the Free Software Foundation.

// BenLincoln.DaVinci is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with BenLincoln.DaVinci (in the file LICENSE.txt).  
// If not, see <http://www.gnu.org/licenses/>.

using System;
using System.Collections;
using System.Collections.Generic;
using System.IO;
using System.Text;
using System.Threading;

using BenLincoln.Utility;

namespace BenLincoln.DaVinci
{
    public struct EngineModule
    {
        public TMSBEngine Engine;
        public bool Committed;
    }

    public struct EngineCoreModule
    {
        public TMSBEngineCore EngineCore;
        public bool Committed;
    }

    public class TMSBEngineController
    {
        protected TMSBEngineControllerConfiguration _EngineControllerConfiguration;
        protected string[] _BaseDirectories;
        //protected const int _MaximumDirectoryRecursion = 128;
        protected const int _SleepLengthWaitingForAvailableEngine = 5000;

        protected TMSBEngineState _CurrentState;

        protected TMSBEngine.DirectoryProcessingStartedHandler _DirectoryProcessingStartedHandler;
        protected TMSBEngine.CubeScriptBeingGeneratedHandler _CubeScriptBeingGeneratedHandler;
        protected TMSBEngine.CubeScriptGeneratedHandler _CubeScriptGeneratedHandler;
        protected TMSBEngine.ImageStartedHandler _EngineImageStartedHandler;
        protected TMSBEngine.ImageFinishedHandler _EngineImageFinishedHandler;
        protected TMSBEngine.EngineExecutionCompleteHandler _EngineExecutionCompleteHandler;
        protected TMSBEngine.ScriptExecutionErrorHandler _ScriptExecutionErrorHandler;
        protected TMSBEngine.ScriptExecutionCompleteHandler _ScriptExecutionCompleteHandler;
        protected TMSBEngine.ScriptExecutionDebugMessageHandler _ScriptExecutionDebugMessageHandler;
        protected TMSBEngine.ScriptExecutionInformationalMessageHandler _ScriptExecutionInformationalMessageHandler;
        protected TMSBEngine.ScriptExecutionAbortedHandler _ScriptExecutionAbortedHandler;
        protected TMSBEngine.DirectoryProcessingFinishedHandler _DirectoryProcessingFinishedHandler;
        protected TMSBEngine.WaitingForProcessingThreadHandler _WaitingForProcessingThreadHandler;
        protected TMSBEngine.ObtainedProcessingThreadHandler _ObtainedProcessingThreadHandler;
        protected TMSBEngine.UnhandledExceptionHandler _UnhandledExceptionHandler;

        protected EngineModule[] _Engines;
        protected EngineCoreModule[] _EngineCores;
        protected NumericValueTracker _MemoryUseTracker;
        protected NumericValueTracker _ProcessorTimeTracker;

        public delegate void WaitingForEngineThreadHandler(object sender, EngineManagementEventArgs e);
        public delegate void ObtainedEngineThreadHandler(object sender, EngineManagementEventArgs e);
        public delegate void EngineControllerExecutionErrorHandler(object sender, EngineManagementEventArgs e);
        public delegate void EngineControllerInformationalMessageHandler(object sender, EngineManagementEventArgs e);
        public delegate void EngineControllerDebugMessageHandler(object sender, EngineManagementEventArgs e);
        public delegate void EngineControllerExecutionStartedHandler(object sender, EngineManagementEventArgs e);
        public delegate void EngineControllerExecutionCompleteHandler(object sender, EngineManagementEventArgs e);

        public event WaitingForEngineThreadHandler WaitingForEngineThread;
        public event ObtainedEngineThreadHandler ObtainedEngineThread;
        public event EngineControllerExecutionErrorHandler EngineControllerExecutionError;
        public event EngineControllerInformationalMessageHandler EngineControllerInformationalMessage;
        public event EngineControllerDebugMessageHandler EngineControllerDebugMessage;
        public event EngineControllerExecutionStartedHandler EngineControllerExecutionStarted;
        public event EngineControllerExecutionCompleteHandler EngineControllerExecutionComplete;

        protected static readonly object EngineThreadMonitorLockObject = new object();
        protected static readonly object EngineCoreStatusLockObject = new object();
        protected static readonly object ProcessingThreadMonitorLockObject = new object();
        protected static readonly object ProcessingThreadRegistrationLockObject = new object();

        #region Properties

        public TMSBEngineState CurrentState
        {
            get
            {
                return _CurrentState;
            }
        }

        public TMSBEngineControllerConfiguration EngineControllerConfiguration
        {
            get
            {
                return _EngineControllerConfiguration;
            }
            set
            {
                _EngineControllerConfiguration = value;
                _Engines = new EngineModule[_EngineControllerConfiguration.EngineThreadCount];
                for (int i = 0; i < _Engines.Length; i++)
                {
                    _Engines[i].Committed = false;
                }
                _BaseDirectories = EngineControllerConfiguration.InputDirectories;
                if (EngineControllerConfiguration.ScanInputDirectoriesRecursively)
                {
                    _BaseDirectories = GetRecursiveDirectoryList(_BaseDirectories);
                }
            }
        }

        public EngineModule[] Engines
        {
            get
            {
                return _Engines;
            }
            set
            {
                _Engines = value;
            }
        }

        public string[] BaseDirectories
        {
            get
            {
                return _BaseDirectories;
            }
        }

        public TMSBEngine.DirectoryProcessingStartedHandler DirectoryProcessingStartedHandler
        {
            get
            {
                return _DirectoryProcessingStartedHandler;
            }
            set
            {
                _DirectoryProcessingStartedHandler = value;
            }
        }

        public TMSBEngine.CubeScriptBeingGeneratedHandler CubeScriptBeingGeneratedHandler
        {
            get
            {
                return _CubeScriptBeingGeneratedHandler;
            }
            set
            {
                _CubeScriptBeingGeneratedHandler = value;
            }
        }

        public TMSBEngine.CubeScriptGeneratedHandler CubeScriptGeneratedHandler
        {
            get
            {
                return _CubeScriptGeneratedHandler;
            }
            set
            {
                _CubeScriptGeneratedHandler = value;
            }
        }

        public TMSBEngine.ImageStartedHandler EngineImageStartedHandler
        {
            get
            {
                return _EngineImageStartedHandler;
            }
            set
            {
                _EngineImageStartedHandler = value;
            }
        }

        public TMSBEngine.ImageFinishedHandler EngineImageFinishedHandler
        {
            get
            {
                return _EngineImageFinishedHandler;
            }
            set
            {
                _EngineImageFinishedHandler = value;
            }
        }

        public TMSBEngine.EngineExecutionCompleteHandler EngineExecutionCompleteHandler
        {
            get
            {
                return _EngineExecutionCompleteHandler;
            }
            set
            {
                _EngineExecutionCompleteHandler = value;
            }
        }

        public TMSBEngine.ScriptExecutionErrorHandler ScriptExecutionErrorHandler
        {
            get
            {
                return _ScriptExecutionErrorHandler;
            }
            set
            {
                _ScriptExecutionErrorHandler = value;
            }
        }

        public TMSBEngine.ScriptExecutionCompleteHandler ScriptExecutionCompleteHandler
        {
            get
            {
                return _ScriptExecutionCompleteHandler;
            }
            set
            {
                _ScriptExecutionCompleteHandler = value;
            }
        }

        public TMSBEngine.ScriptExecutionDebugMessageHandler ScriptExecutionDebugMessageHandler
        {
            get
            {
                return _ScriptExecutionDebugMessageHandler;
            }
            set
            {
                _ScriptExecutionDebugMessageHandler = value;
            }
        }

        public TMSBEngine.ScriptExecutionInformationalMessageHandler ScriptExecutionInformationalMessageHandler
        {
            get
            {
                return _ScriptExecutionInformationalMessageHandler;
            }
            set
            {
                _ScriptExecutionInformationalMessageHandler = value;
            }
        }

        public TMSBEngine.ScriptExecutionAbortedHandler ScriptExecutionAbortedHandler
        {
            get
            {
                return _ScriptExecutionAbortedHandler;
            }
            set
            {
                _ScriptExecutionAbortedHandler = value;
            }
        }

        public TMSBEngine.DirectoryProcessingFinishedHandler DirectoryProcessingFinishedHandler
        {
            get
            {
                return _DirectoryProcessingFinishedHandler;
            }
            set
            {
                _DirectoryProcessingFinishedHandler = value;
            }
        }

        public TMSBEngine.WaitingForProcessingThreadHandler WaitingForProcessingThreadHandler
        {
            get
            {
                return _WaitingForProcessingThreadHandler;
            }
            set
            {
                _WaitingForProcessingThreadHandler = value;
            }
        }

        public TMSBEngine.ObtainedProcessingThreadHandler ObtainedProcessingThreadHandler
        {
            get
            {
                return _ObtainedProcessingThreadHandler;
            }
            set
            {
                _ObtainedProcessingThreadHandler = value;
            }
        }

        public TMSBEngine.UnhandledExceptionHandler UnhandledExceptionHandler
        {
            get
            {
                return _UnhandledExceptionHandler;
            }
            set
            {
                _UnhandledExceptionHandler = value;
            }
        }

        public NumericValueTracker MemoryUseTracker
        {
            get
            {
                return _MemoryUseTracker;
            }
        }

        public NumericValueTracker ProcessorTimeTracker
        {
            get
            {
                return _ProcessorTimeTracker;
            }
        }

        #endregion

        public TMSBEngineController(TMSBEngineControllerConfiguration ecConfig)
        {
            EngineControllerConfiguration = ecConfig;
            _Engines = new EngineModule[EngineControllerConfiguration.EngineThreadCount];
            _EngineCores = new EngineCoreModule[EngineControllerConfiguration.GlobalThreadCount];
            _MemoryUseTracker = new NumericValueTracker();
            _ProcessorTimeTracker = new NumericValueTracker();

            _CurrentState = TMSBEngineState.Running;
        }

        ~TMSBEngineController()
        {
            foreach (EngineModule engine in _Engines)
            {
                if (engine.Engine != null)
                {
                    engine.Engine.Stop();
                }
            }
            foreach (EngineCoreModule core in _EngineCores)
            {
                if (core.EngineCore != null)
                {
                    core.EngineCore.KillThreads();
                }
            }
            GC.Collect();
        }

        public static string GetUnsupportedOSMessage(string osName)
        {
            return "The operating system/platform '" + osName +
                            "' is not currently supported.";
        }

        public void ExecuteProcessingThreaded()
        {
            Thread controllerThread = new Thread(new ThreadStart(ExecuteProcessing));
            controllerThread.Start();
        }

        public void ExecuteProcessing()
        {
            GeneratedEngineControllerDebugEvent(EngineControllerConfiguration.GetConfigurationInformationString());

            string beginMessage = "Beginning processing of " + BaseDirectories.Length.ToString();
            if (BaseDirectories.Length == 1)
            {
                beginMessage += " directory.";
            }
            else
            {
                beginMessage += " directories.";
            }
            GenerateEngineControllerExecutionStartedEvent(beginMessage);

            foreach (string baseDirectory in BaseDirectories)
            {
                int nextThreadNumber = -1;
                DateTime waitStart = DateTime.Now;
                bool hasWaited = false;
                if (CurrentState == TMSBEngineState.Stopped)
                {
                    // do nothing - let the loop finish iterating without doing anything
                    Thread.Sleep(_SleepLengthWaitingForAvailableEngine);
                }
                else
                {
                    while (CurrentState == TMSBEngineState.Paused)
                    {
                        Thread.Sleep(_SleepLengthWaitingForAvailableEngine);
                    }

                    Random rand = new Random();

                    while (nextThreadNumber == -1)
                    {
                        nextThreadNumber = GetNextAvailableEngineThreadNumber();
                        TimeSpan waitTime = DateTime.Now - waitStart;
                        if (nextThreadNumber == -1)
                        {
                            string waitMessage = "Waiting for an available engine to assign to the directory '" + baseDirectory +
                                "'.";
                            if (hasWaited)
                            {
                                waitMessage += " Elapsed wait time: " + TimeTracker.GetLongFormattedTimeSpanString(waitTime);
                            }
                            GenerateWaitingForEngineThreadEvent(waitMessage);
                            Thread.Sleep(_SleepLengthWaitingForAvailableEngine);
                            hasWaited = true;
                        }
                        else
                        {
                            string waitMessage = "Assigned engine number " + nextThreadNumber.ToString() + " to the directory '" +
                                baseDirectory + "'.";
                            if (hasWaited)
                            {
                                waitMessage += " Elapsed wait time: " + TimeTracker.GetLongFormattedTimeSpanString(waitTime);
                            }
                            GenerateObtainedEngineThreadEvent(waitMessage);
                            hasWaited = false;
                            break;
                        }
                        // wait a random number of milliseconds to make simultaneous actions less likely
                        Thread.Sleep(rand.Next(200));
                    }

                    if (CurrentState == TMSBEngineState.Stopped)
                    {
                        GeneratedEngineControllerDebugEvent("An attempt was made to assign a directory to an engine after the engine controller received a stop command. " +
                            "This is probably due to the controller being in a wait state when the stop command was issued.");
                    }
                    else
                    {
                        Engines[nextThreadNumber].Engine = new TMSBEngine(this, nextThreadNumber, baseDirectory, EngineControllerConfiguration);
                        if (DirectoryProcessingStartedHandler != null)
                        {
                            Engines[nextThreadNumber].Engine.DirectoryProcessingStarted += DirectoryProcessingStartedHandler;
                        }
                        if (CubeScriptBeingGeneratedHandler != null)
                        {
                            Engines[nextThreadNumber].Engine.CubeScriptBeingGenerated += CubeScriptBeingGeneratedHandler;
                        }
                        if (CubeScriptGeneratedHandler != null)
                        {
                            Engines[nextThreadNumber].Engine.CubeScriptGenerated += CubeScriptGeneratedHandler;
                        }
                        if (EngineImageStartedHandler != null)
                        {
                            Engines[nextThreadNumber].Engine.ImageStarted += EngineImageStartedHandler;
                        }
                        if (EngineImageFinishedHandler != null)
                        {
                            Engines[nextThreadNumber].Engine.ImageFinished += EngineImageFinishedHandler;
                        }
                        if (EngineExecutionCompleteHandler != null)
                        {
                            Engines[nextThreadNumber].Engine.EngineExecutionComplete += EngineExecutionCompleteHandler;
                        }
                        if (ScriptExecutionCompleteHandler != null)
                        {
                            Engines[nextThreadNumber].Engine.ScriptExecutionComplete += ScriptExecutionCompleteHandler;
                        }
                        if (ScriptExecutionErrorHandler != null)
                        {
                            Engines[nextThreadNumber].Engine.ScriptExecutionError += ScriptExecutionErrorHandler;
                        }
                        if (ScriptExecutionAbortedHandler != null)
                        {
                            Engines[nextThreadNumber].Engine.ScriptExecutionAborted += ScriptExecutionAbortedHandler;
                        }
                        if (ScriptExecutionDebugMessageHandler != null)
                        {
                            Engines[nextThreadNumber].Engine.ScriptExecutionDebugMessageGenerated += ScriptExecutionDebugMessageHandler;
                        }
                        if (ScriptExecutionInformationalMessageHandler != null)
                        {
                            Engines[nextThreadNumber].Engine.ScriptExecutionInformationalMessageGenerated += ScriptExecutionInformationalMessageHandler;
                        }
                        if (DirectoryProcessingFinishedHandler != null)
                        {
                            Engines[nextThreadNumber].Engine.DirectoryProcessingFinished += DirectoryProcessingFinishedHandler;
                        }
                        if (WaitingForProcessingThreadHandler != null)
                        {
                            Engines[nextThreadNumber].Engine.WaitingForProcessingThread += WaitingForProcessingThreadHandler;
                        }
                        if (ObtainedProcessingThreadHandler != null)
                        {
                            Engines[nextThreadNumber].Engine.ObtainedProcessingThread += ObtainedProcessingThreadHandler;
                        }
                        if (UnhandledExceptionHandler != null)
                        {
                            Engines[nextThreadNumber].Engine.UnhandledExceptionThrown += UnhandledExceptionHandler;
                        }
                        Engines[nextThreadNumber].Engine.DirectoryProcessingFinished += OnDirectoryProcessingFinished;
                        Engines[nextThreadNumber].Engine.EngineExecutionComplete += OnEngineExecutionComplete;
                        Engines[nextThreadNumber].Engine.ImageFinished += OnImageFinished;

                        Engines[nextThreadNumber].Engine.BaseDirectory = baseDirectory;
                        Engines[nextThreadNumber].Engine.EngineControllerConfig = _EngineControllerConfiguration;

                        Engines[nextThreadNumber].Engine.ExecuteProcessingThreaded();
                    }
                }

            }
            bool waitingForEngines = true;
            while (waitingForEngines)
            {
                bool foundActiveEngine = false;
                foreach (EngineModule engine in _Engines)
                {
                    if (engine.Committed == true)
                    {
                        foundActiveEngine = true;
                    }
                    if ((engine.Engine != null) && (engine.Engine.CurrentState != TMSBEngineState.Finished) &&
                        (engine.Engine.CurrentState != TMSBEngineState.Stopped))
                    {
                        foundActiveEngine = true;
                    }
                }
                if (!foundActiveEngine)
                {
                    waitingForEngines = false;
                }
                else
                {
                    Thread.Sleep(_SleepLengthWaitingForAvailableEngine);
                }
            }

            string endMessage = "Finished processing " + BaseDirectories.Length.ToString();
            if (BaseDirectories.Length == 1)
            {
                endMessage += " directory.";
            }
            else
            {
                endMessage += " directories.";
            }
            GenerateEngineControllerExecutionCompleteEvent(endMessage);
        }

        public void OnDirectoryProcessingFinished(object sender, DirectoryProcessingArgs d)
        {
            //Engines[d.ThreadNumber].Committed = false;
        }

        //public void OnEngineExecutionComplete(object sender, DirectoryProcessingArgs d)
        //{
        //    Engines[d.ThreadNumber].Committed = false;
        //}

        public void GenerateProcessingStatisticsMessage()
        {
            // gathering this data is relatively involved, so only capture if it it will be used
            //if (EngineControllerDebugMessage != null)
            if (EngineControllerDebugMessage != null)
            {
                // split into three messages for readability

                // message 1: overall stats
                StringBuilder builder = new StringBuilder();
                int activeEngines = 0;
                int[] activeThreadsPerEngine = new int[0];
                int totalActiveThreads = GetActiveProcessingThreadCount(out activeEngines, out activeThreadsPerEngine);
                builder.Append("Active Engines: ");
                builder.Append(activeEngines.ToString());
                builder.Append("/");
                builder.Append(EngineControllerConfiguration.EngineThreadCount.ToString());
                builder.Append(". Total Active Processing Threads: ");
                builder.Append(totalActiveThreads);
                builder.Append("/");
                builder.Append(EngineControllerConfiguration.GlobalThreadCount.ToString());
                builder.Append(".");
                EngineManagementEventArgs e = new EngineManagementEventArgs(-1, -1, builder.ToString());
                EngineControllerDebugMessage(this, e);
                //EngineControllerInformationalMessage(this, e);

                // message 2: per-engine thread count
                builder = new StringBuilder();
                builder.Append("Processing Thread Count By Engine Number: ");
                for (int i = 0; i < activeThreadsPerEngine.Length; i++)
                {
                    builder.Append(String.Format("[E{0:00}/", i));
                    if (activeThreadsPerEngine[i] >= 0)
                    {
                        builder.Append(String.Format("{0:00}", activeThreadsPerEngine[i]));
                    }
                    else
                    {
                        builder.Append("Inactive");
                    }
                    builder.Append("]");
                    if (i < (activeThreadsPerEngine.Length - 1))
                    {
                        builder.Append(", ");
                    }
                }
                e = new EngineManagementEventArgs(-1, -1, builder.ToString());
                EngineControllerDebugMessage(this, e);
                //EngineControllerInformationalMessage(this, e);

                if (!EngineControllerConfiguration.CollectPerformanceData)
                {
                    return;
                }

                // message 3: per-thread memory use
                builder = new StringBuilder();
                builder.Append("Memory Usage By Thread Number: ");
                for (int i = 0; i < _EngineCores.Length; i++)
                {
                    builder.Append(String.Format("[T{0:00}/", i));
                    if ((_EngineCores[i].EngineCore == null) || (!_EngineCores[i].Committed))
                    {
                        builder.Append("Inactive");
                    }
                    else
                    {
                        double currentMemoryUse = _EngineCores[i].EngineCore.MemoryUseTracker.CurrentValue;
                        if (currentMemoryUse == _EngineCores[i].EngineCore.MemoryUseTracker.DefaultValue)
                        {
                            currentMemoryUse = 0;
                        }
                        builder.Append(GetMemoryUsageString(currentMemoryUse));
                    }
                    builder.Append("]");
                }

                e = new EngineManagementEventArgs(-1, -1, builder.ToString());
                EngineControllerDebugMessage(this, e);
                //EngineControllerInformationalMessage(this, e);


                // message 4: per-thread % processor time use
                builder = new StringBuilder();
                builder.Append("Percent Processor Time Usage By Thread Number: ");
                for (int i = 0; i < _EngineCores.Length; i++)
                {
                    builder.Append(String.Format("[T{0:00}/", i));
                    if ((_EngineCores[i].EngineCore == null) || (!_EngineCores[i].Committed))
                    {
                        builder.Append("Inactive");
                    }
                    else
                    {
                        double currentPPTUse = _EngineCores[i].EngineCore.ProcessorTimeTracker.CurrentValue;
                        if (currentPPTUse == _EngineCores[i].EngineCore.ProcessorTimeTracker.DefaultValue)
                        {
                            currentPPTUse = 0;
                        }
                        builder.Append(GetPercentProcessorTimeString(currentPPTUse));
                    }
                    builder.Append("]");
                }

                e = new EngineManagementEventArgs(-1, -1, builder.ToString());
                EngineControllerDebugMessage(this, e);
                //EngineControllerInformationalMessage(this, e);
            }
        }

        public string GetMemoryUsageString(double value)
        {
            return GetMemoryUsageString(value, false);
        }

        public string GetMemoryUsageString(double value, bool returnEmptyStringIfNoData)
        {
            if (returnEmptyStringIfNoData && value < 0)
            {
                return "";
            }
            if (_EngineControllerConfiguration.MemoryUnits == MemoryRepresentation.PowerOfTwo)
            {
                return NumericValueTracker.GetValueAsBytesInPowerOfTwoStringFormat(value);
            }
            else
            {
                return NumericValueTracker.GetValueAsBytesInPowerOfTenStringFormat(value);
            }
        }

        public string GetPercentProcessorTimeString(double value)
        {
            return GetPercentProcessorTimeString(value, false);
        }

        public string GetPercentProcessorTimeString(double value, bool returnEmptyStringIfNoData)
        {
            if (returnEmptyStringIfNoData && value < 0)
            {
                return "";
            }
            if (value < 10)
            {
                return String.Format("{0:0.0}", value) + "%";
            }
            return String.Format("{0:00}", value) + "%";
        }

        protected int GetNextAvailableEngineThreadNumber()
        {
            int result = -1;
            lock (EngineThreadMonitorLockObject)
            {
                for (int i = 0; i < _Engines.Length; i++)
                {
                    if (_Engines[i].Engine == null)
                    {
                        _Engines[i] = new EngineModule();
                    }
                    if (!_Engines[i].Committed)
                    {
                        _Engines[i].Committed = true;
                        result = i;
                        break;
                    }
                }
            }

            GenerateProcessingStatisticsMessage();

            return result;
        }

        public int GetActiveEngineCount()
        {
            int activeEngines = 0;
            for (int i = 0; i < _Engines.Length; i++)
            {
                if (_Engines[i].Engine != null)
                {
                    if ((_Engines[i].Engine.CurrentState == TMSBEngineState.Running)
                        ||
                        (_Engines[i].Engine.CurrentState == TMSBEngineState.Pausing)
                        ||
                        (_Engines[i].Engine.CurrentState == TMSBEngineState.Stopping))
                    {
                        activeEngines++;
                    }
                }
            }
            return activeEngines;
        }

        public int GetActiveProcessingThreadCount(out int activeEngines, out int[] activeThreadsPerEngine)
        {
            int total = 0;
            activeEngines = 0;
            activeThreadsPerEngine = new int[_Engines.Length];
            for (int i = 0; i < _Engines.Length; i++)
            {
                if (_Engines[i].Engine != null)
                {
                    total += _Engines[i].Engine.ActiveEngineCoreCount;
                    activeThreadsPerEngine[i] = _Engines[i].Engine.ActiveEngineCoreCount;
                    if ((_Engines[i].Engine.CurrentState == TMSBEngineState.Running)
                        ||
                        (_Engines[i].Engine.CurrentState == TMSBEngineState.Pausing)
                        ||
                        (_Engines[i].Engine.CurrentState == TMSBEngineState.Stopping))
                    {
                        activeEngines++;
                    }
                }
                else
                {
                    activeThreadsPerEngine[i] = -1;
                }
            }
            return total;
        }

        public int GetNextAvailableProcessingThreadNumber()
        {
            int result = -1;
            lock (ProcessingThreadMonitorLockObject)
            {
                UpdateEngineCoreStatus(); 

                int activeEngines;
                int[] activeThreadsPerEngine;
                int activeThreads = GetActiveProcessingThreadCount(out activeEngines, out activeThreadsPerEngine);

                if (activeThreads >= _EngineCores.Length)
                {
                    return -1;
                }

                for (int i = 0; i < _EngineCores.Length; i++)
                {
                    if ((_EngineCores[i].EngineCore == null) || (!_EngineCores[i].Committed))
                    {
                        _EngineCores[i].Committed = true;
                        result = i;
                        GeneratedEngineControllerDebugEvent("Assigned EngineCore number " +
                            i.ToString() + " to a requesting Engine.");
                        break;
                    }
                }
            }

            GenerateProcessingStatisticsMessage();

            return result;
        }

        public void UpdateEngineCoreStatus()
        {
            //lock (EngineCoreStatusLockObject)
            //{
                for (int i = 0; i < _EngineCores.Length; i++)
                {
                    if (_EngineCores[i].EngineCore != null)
                    {
                        if (_EngineCores[i].EngineCore.FinishedExecuting)
                        {
                            _EngineCores[i].Committed = false;
                            GeneratedEngineControllerDebugEvent("Found the EngineCore object numbered " +
                                i.ToString() + " in a FinishedExecuting state, and marked it as uncommitted.");
                        }
                    }
                }
            //}

            UpdateEngineCorePerformanceData();
        }

        protected void UpdateEngineCorePerformanceData()
        {
            if (!_EngineControllerConfiguration.CollectPerformanceData)
            {
                return;
            }
            double totalMemoryUse = 0.0d;
            for (int i = 0; i < _Engines.Length; i++)
            {
                if ((_Engines[i].Engine != null) && (_Engines[i].Engine.ActiveEngineCoreCount > 0))
                {
                    try
                    {
                        _Engines[i].Engine.UpdateEngineCoreStatus();
                        double mem = _Engines[i].Engine.MemoryUseTracker.CurrentValue;
                        totalMemoryUse += mem;
                        //_MemoryUseTracker.UpdateMaximum(_Engines[i].Engine.MemoryUseTracker.Maximum);
                        //_MemoryUseTracker.UpdateMinimum(_Engines[i].Engine.MemoryUseTracker.Minimum);
                    }
                    catch (Exception ex)
                    {
                        GeneratedEngineControllerDebugEvent("Unable to obtain the current memory in use by Engine number "
                            + i.ToString() + ". The error description was '" + ex.Message + "'.");
                    }
                }
            }
            _MemoryUseTracker.AddValue(totalMemoryUse);

            double totalPPT = 0.0d;
            for (int i = 0; i < _Engines.Length; i++)
            {
                if ((_Engines[i].Engine != null) && (_Engines[i].Engine.ActiveEngineCoreCount > 0))
                {
                    try
                    {
                        double ppt = _Engines[i].Engine.ProcessorTimeTracker.CurrentValue;
                        totalPPT += ppt;
                        //_ProcessorTimeTracker.UpdateMaximum(_Engines[i].Engine.ProcessorTimeTracker.Maximum);
                        //_ProcessorTimeTracker.UpdateMinimum(_Engines[i].Engine.ProcessorTimeTracker.Minimum);
                    }
                    catch (Exception ex)
                    {
                        GeneratedEngineControllerDebugEvent("Unable to obtain the current percent processor time in use by Engine number "
                            + i.ToString() + ". The error description was '" + ex.Message + "'.");
                    }
                }
            }
            _ProcessorTimeTracker.AddValue(totalPPT);
        }

        public void RegisterEngineCore(int coreNumber, ref TMSBEngineCore core)
        {
            lock (ProcessingThreadRegistrationLockObject)
            {
                if ((_EngineCores[coreNumber].EngineCore != null) && (!_EngineCores[coreNumber].Committed) && (!_EngineCores[coreNumber].EngineCore.FinishedExecuting))
                //if ((_EngineCores[coreNumber].EngineCore != null) && (!_EngineCores[coreNumber].Committed))
                {
                    StringBuilder builder = new StringBuilder();
                    builder.Append("An attempt was made to register an EngineCore object under the number ");
                    builder.Append(coreNumber.ToString());
                    builder.Append(", but this number was already marked as registered to another active object. ");
                    builder.Append("The existing EngineCore returned this descriptive information: [");
                    builder.Append(_EngineCores[coreNumber].EngineCore.GetProcessingInformationString());
                    builder.Append("], and the new EngineCore returned this descriptive information: [");
                    builder.Append(core.GetProcessingInformationString());
                    builder.Append("].");
                    throw new Exception(builder.ToString());
                }
                GeneratedEngineControllerDebugEvent("Registered as EngineCore number " + coreNumber.ToString() +
                    " the EngineCore object with this descriptive information: [" +
                    core.GetProcessingInformationString() + "].");
                _EngineCores[coreNumber].EngineCore = core;
                _EngineCores[coreNumber].Committed = true;
            }
        }

        public void OnEngineExecutionComplete(object sender, DirectoryProcessingArgs d)
        {
            if (DirectoryProcessingStartedHandler != null)
            {
                _Engines[d.EngineNumber].Engine.DirectoryProcessingStarted -= DirectoryProcessingStartedHandler;
            }
            if (CubeScriptBeingGeneratedHandler != null)
            {
                _Engines[d.EngineNumber].Engine.CubeScriptBeingGenerated -= CubeScriptBeingGeneratedHandler;
            }
            if (CubeScriptGeneratedHandler != null)
            {
                _Engines[d.EngineNumber].Engine.CubeScriptGenerated -= CubeScriptGeneratedHandler;
            }
            if (EngineImageStartedHandler != null)
            {
                _Engines[d.EngineNumber].Engine.ImageStarted -= EngineImageStartedHandler;
            }
            if (EngineImageFinishedHandler != null)
            {
                _Engines[d.EngineNumber].Engine.ImageFinished -= EngineImageFinishedHandler;
            }
            if (EngineExecutionCompleteHandler != null)
            {
                _Engines[d.EngineNumber].Engine.EngineExecutionComplete -= EngineExecutionCompleteHandler;
            }
            if (ScriptExecutionCompleteHandler != null)
            {
                _Engines[d.EngineNumber].Engine.ScriptExecutionComplete -= ScriptExecutionCompleteHandler;
            }
            if (ScriptExecutionErrorHandler != null)
            {
                _Engines[d.EngineNumber].Engine.ScriptExecutionError -= ScriptExecutionErrorHandler;
            }
            if (ScriptExecutionAbortedHandler != null)
            {
                _Engines[d.EngineNumber].Engine.ScriptExecutionAborted -= ScriptExecutionAbortedHandler;
            }
            if (ScriptExecutionDebugMessageHandler != null)
            {
                _Engines[d.EngineNumber].Engine.ScriptExecutionDebugMessageGenerated -= ScriptExecutionDebugMessageHandler;
            }
            if (ScriptExecutionInformationalMessageHandler != null)
            {
                _Engines[d.EngineNumber].Engine.ScriptExecutionInformationalMessageGenerated -= ScriptExecutionInformationalMessageHandler;
            }
            if (DirectoryProcessingFinishedHandler != null)
            {
                _Engines[d.EngineNumber].Engine.DirectoryProcessingFinished -= DirectoryProcessingFinishedHandler;
            }
            if (WaitingForProcessingThreadHandler != null)
            {
                _Engines[d.EngineNumber].Engine.WaitingForProcessingThread -= WaitingForProcessingThreadHandler;
            }
            if (ObtainedProcessingThreadHandler != null)
            {
                _Engines[d.EngineNumber].Engine.ObtainedProcessingThread -= ObtainedProcessingThreadHandler;
            }
            if (UnhandledExceptionHandler != null)
            {
                _Engines[d.EngineNumber].Engine.UnhandledExceptionThrown -= UnhandledExceptionHandler;
            }
            _Engines[d.EngineNumber].Engine.DirectoryProcessingFinished -= OnDirectoryProcessingFinished;
            _Engines[d.EngineNumber].Engine.EngineExecutionComplete -= OnEngineExecutionComplete;
            _Engines[d.EngineNumber].Engine.ImageFinished -= OnImageFinished;

            _Engines[d.EngineNumber].Committed = false;
        }

        public void OnImageFinished(object sender, ImageFinishedEventArgs i)
        {
            _EngineCores[i.GlobalThreadNumber].Committed = false;
        }

        protected string[] GetRecursiveDirectoryList(string baseDirectory)
        {
            return GetRecursiveDirectoryList(new string[] { baseDirectory });
        }

        protected string[] GetRecursiveDirectoryList(string[] baseDirectories)
        {
            DirectoryLister dl = new DirectoryLister();
            string[] errors = new string[0];
            string[] result = dl.GetRecursiveDirectoryList(baseDirectories, out errors);
            if (errors.Length > 0)
            {
                LogError(errors.Length.ToString() + " errors occurred:");
                foreach (string errorMessage in errors)
                {
                    LogError("\t" + errorMessage);
                }
            }
            return result;
        }

        //protected string[] GetRecursiveDirectoryList(string baseDirectory, int currentRecursionLevel)
        //{
        //    return GetRecursiveDirectoryList(new string[] { baseDirectory }, currentRecursionLevel);
        //}

        //protected string[] GetRecursiveDirectoryList(string[] baseDirectories, int currentRecursionLevel)
        //{
        //    ArrayList result = new ArrayList();

        //    foreach (string baseDir in baseDirectories)
        //    {
        //        if (!result.Contains(baseDir))
        //        {
        //            // only add directories to the list if they actually contain at least one file
        //            if (Directory.GetFiles(baseDir).Length > 0)
        //            {
        //                result.Add(baseDir);
        //            }

        //            if (currentRecursionLevel < _MaximumDirectoryRecursion)
        //            {
        //                try
        //                {
        //                    string[] subDirPaths = Directory.GetDirectories(baseDir);

        //                    foreach (string subDirPath in subDirPaths)
        //                    {
        //                        try
        //                        {
        //                            string[] subDirs = GetRecursiveDirectoryList(subDirPath, currentRecursionLevel + 1);
        //                            foreach (string subDir in subDirs)
        //                            {
        //                                result.Add(subDir);
        //                            }
        //                        }
        //                        catch (Exception ex)
        //                        {
        //                                LogError("unable to access the directory '" + baseDir + "'. " +
        //                                    "The error message was '" + ex.Message + "'.");
        //                        }
        //                    }
        //                }
        //                catch (Exception ex)
        //                {
        //                    LogError("unable to scan the directory '" + baseDir + "' for subdirectories. " +
        //                        "The error message was '" + ex.Message + "'.");
        //                }
        //            }
        //            else
        //            {
        //                LogError("the maximum depth of directory recursion was reached with the directory '" + baseDir +
        //                    "'. Directories below this level will not be scanned.");
        //            }
        //        }
        //    }

        //    return (string[])result.ToArray("".GetType());
        //}

        public void GenerateWaitingForEngineThreadEvent(string information)
        {
            if (WaitingForEngineThread != null)
            {
                EngineManagementEventArgs e = new EngineManagementEventArgs(-1, -1, information);
                WaitingForEngineThread(this, e);
            }
        }

        public void GenerateObtainedEngineThreadEvent(string information)
        {
            if (ObtainedEngineThread != null)
            {
                EngineManagementEventArgs e = new EngineManagementEventArgs(-1, -1, information);
                ObtainedEngineThread(this, e);
            }
        }

        public void GenerateEngineExecutionErrorEvent(string information)
        {
            if (EngineControllerExecutionError != null)
            {
                EngineManagementEventArgs e = new EngineManagementEventArgs(-1, -1, information);
                EngineControllerExecutionError(this, e);
            }
        }

        public void GeneratedEngineControllerInformationalEvent(string information)
        {
            if (EngineControllerInformationalMessage != null)
            {
                EngineManagementEventArgs e = new EngineManagementEventArgs(-1, -1, information);
                EngineControllerInformationalMessage(this, e);
            }
        }

        public void GeneratedEngineControllerDebugEvent(string information)
        {
            if (EngineControllerDebugMessage != null)
            {
                EngineManagementEventArgs e = new EngineManagementEventArgs(-1, -1, information);
                EngineControllerDebugMessage(this, e);
            }
        }

        public void GenerateEngineControllerExecutionStartedEvent(string information)
        {
            if (EngineControllerExecutionStarted != null)
            {
                EngineManagementEventArgs e = new EngineManagementEventArgs(-1, -1, information);
                EngineControllerExecutionStarted(this, e);
            }
        }

        public void GenerateEngineControllerExecutionCompleteEvent(string information)
        {
            if (EngineControllerExecutionComplete != null)
            {
                EngineManagementEventArgs e = new EngineManagementEventArgs(-1, -1, information);
                EngineControllerExecutionComplete(this, e);
            }
        }

        protected void LogError(string message)
        {
            // event handler call goes here
            //Console.WriteLine(message);
            GenerateEngineExecutionErrorEvent("Error: " + message);
        }

        public void Pause()
        {
            _CurrentState = TMSBEngineState.Paused;
            for (int i = 0; i < Engines.Length; i++)
            {
                if ((Engines[i].Engine != null) && (Engines[i].Engine.CurrentState == TMSBEngineState.Running))
                {
                    Engines[i].Engine.Pause();
                }
            }
        }

        public void Resume()
        {
            _CurrentState = TMSBEngineState.Running;

            for (int i = 0; i < Engines.Length; i++)
            {
                if ((Engines[i].Engine != null) && 
                    ((Engines[i].Engine.CurrentState == TMSBEngineState.Paused) || (Engines[i].Engine.CurrentState == TMSBEngineState.Pausing)))
                {
                    Engines[i].Engine.Resume();
                }
            }
        }

        public void Stop()
        {
            _CurrentState = TMSBEngineState.Stopped;

            for (int i = 0; i < Engines.Length; i++)
            {
                if ((Engines[i].Engine != null) &&
                    (Engines[i].Engine.CurrentState != TMSBEngineState.Finished))
                {
                    Engines[i].Engine.Stop();
                }
            }
        }

        public void ForceStop()
        {
            _CurrentState = TMSBEngineState.Stopped;
            for (int i = 0; i < Engines.Length; i++)
            {
                if (Engines[i].Engine != null)
                {
                    Engines[i].Engine.ForceStop();
                    Engines[i].Committed = false;
                }
            }
        }
    }
}
