// BenLincoln.DaVinci - A shared library used by The Mirror's Surface Breaks
// Copyright 2010-2011 Ben Lincoln
// http://www.beneaththewaves.net/
//

// This file is part of BenLincoln.DaVinci.

// BenLincoln.DaVinci is free software: you can redistribute it and/or modify
// it under the terms of version 3 of the GNU General Public License as published by
// the Free Software Foundation.

// BenLincoln.DaVinci is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with BenLincoln.DaVinci (in the file LICENSE.txt).  
// If not, see <http://www.gnu.org/licenses/>.

using System;
using System.Collections.Generic;
using System.IO;
using System.Text;
using System.Threading;

using BenLincoln.Utility;

namespace BenLincoln.DaVinci
{
    public class TMSBGUIEngineMonitor : TMSBEngineMonitor, IDisposable
    {
        public delegate void EngineControllerExecutionCompleteHandler(object sender, EngineManagementEventArgs e);

        public delegate void EngineExecutionCompleteHandler(object sender, DirectoryProcessingArgs d);

        public delegate void DirectoryProcessingStartedHandler(object sender, DirectoryProcessingArgs d);
        public delegate void DirectoryProcessingErrorHandler(object sender, DirectoryProcessingArgs d);
        public delegate void DirectoryProcessingFinishedHandler(object sender, DirectoryProcessingArgs d);

        public delegate void ImageProcessingStartedHandler(object sender, ImageEventArgs a);
        public delegate void ImageProcessingFinishedHandler(object sender, ImageFinishedEventArgs a);

        public delegate void ScriptBeingGeneratedHandler(object sender, ThreadedEventArgs t);
        public delegate void ScriptGeneratedHandler(object sender, CubeScriptGeneratedEventArgs t);

        public delegate void WaitingForProcessingThreadHandler(object sender, EngineManagementEventArgs e);


        public event EngineControllerExecutionCompleteHandler EngineControllerExecutionComplete;

        public event EngineExecutionCompleteHandler EngineExecutionComplete;

        public event DirectoryProcessingStartedHandler DirectoryProcessingStarted;
        public event DirectoryProcessingErrorHandler DirectoryProcessingError;
        public event DirectoryProcessingFinishedHandler DirectoryProcessingFinished;

        public event ImageProcessingStartedHandler ImageProcessingStarted;
        public event ImageProcessingFinishedHandler ImageProcessingFinished;

        public event ScriptBeingGeneratedHandler ScriptBeingGenerated;
        public event ScriptGeneratedHandler ScriptGenerated;

        public event WaitingForProcessingThreadHandler WaitingForProcessingThread;

        public TMSBGUIEngineMonitor(TMSBEngineController controller, OutputLevel outputLevel)
            : base(controller, outputLevel)
        {
            // locked so that all event handlers are registered before logging begins
            lock (GlobalLock.LockObject)
            {
                _MetaTracker.TotalElementCount = controller.BaseDirectories.Length;

                //controller.WaitingForEngineThread += OnGeneralThreadManagementEvent;
                //controller.ObtainedEngineThread += OnGeneralThreadManagementEvent;
                //controller.WaitingForProcessingThreadHandler += OnGeneralThreadManagementEvent;
                //controller.ObtainedProcessingThreadHandler += OnGeneralThreadManagementEvent;

                controller.EngineImageStartedHandler += OnImageStarted;
                controller.EngineImageFinishedHandler += OnImageFinished;

                controller.EngineExecutionCompleteHandler += OnEngineExecutionComplete;

                controller.DirectoryProcessingStartedHandler += OnDirectoryProcessingStarted;
                //controller.ScriptExecutionCompleteHandler += OnScriptExecutionComplete;
                controller.DirectoryProcessingFinishedHandler += OnDirectoryProcessingFinished;
                //controller.EngineControllerExecutionStarted += OnEngineControllerExecutionStarted;
                controller.EngineControllerExecutionComplete += OnEngineControllerExecutionComplete;

                controller.CubeScriptGeneratedHandler += OnScriptGenerated;

                controller.WaitingForProcessingThreadHandler += OnWaitingForProcessingThread;

                controller.ScriptExecutionAbortedHandler += OnScriptExecutionAborted;
                controller.ScriptExecutionErrorHandler += OnScriptExecutionError;
                controller.UnhandledExceptionHandler += OnUnhandledExceptionThrown;
            }
        }

        protected override void Dispose(bool disposing)
        {
            // If you need thread safety, use a lock around these 
            // operations, as well as in your methods that use the resource.
            if (!_Disposed)
            {
                //if (disposing)
                //{
                    _Controller.EngineImageStartedHandler -= OnImageStarted;
                    _Controller.EngineImageFinishedHandler -= OnImageFinished;

                    _Controller.EngineExecutionCompleteHandler -= OnEngineExecutionComplete;

                    _Controller.DirectoryProcessingStartedHandler -= OnDirectoryProcessingStarted;
                    _Controller.DirectoryProcessingFinishedHandler -= OnDirectoryProcessingFinished;
                    _Controller.EngineControllerExecutionComplete -= OnEngineControllerExecutionComplete;

                    _Controller.CubeScriptGeneratedHandler -= OnScriptGenerated;

                    _Controller.WaitingForProcessingThreadHandler -= OnWaitingForProcessingThread;

                    _Controller.ScriptExecutionAbortedHandler -= OnScriptExecutionAborted;
                    _Controller.ScriptExecutionErrorHandler -= OnScriptExecutionError;
                    _Controller.UnhandledExceptionHandler -= OnUnhandledExceptionThrown;
                //}

                //_Disposed = true;
            }
            base.Dispose(disposing);
        }

        public void OnGeneralThreadManagementEvent(object sender, EngineManagementEventArgs e)
        {
            //WriteOutput(e.EngineNumber, e.GlobalThreadNumber, e.Message);
        }

        public void OnEngineControllerExecutionError(object sender, EngineManagementEventArgs e)
        {
            //WriteOutput(e.EngineNumber, e.GlobalThreadNumber, e.Message);
        }

        protected TimeSpan GetElapsedTimeForEngineNumber(int engineNumber)
        {
            TimeSpan elapsed = TimeSpan.FromSeconds(1.0);
            if (engineNumber > -1)
            {
                if (_Controller.Engines[engineNumber].Engine != null)
                {
                    elapsed = _Controller.Engines[engineNumber].Engine.Tracker.GetTotalElapsedTime();
                }
            }
            return elapsed;
        }

        public void OnUnhandledExceptionThrown(object sender, UnhandledExceptionArgs u)
        {
            if (DirectoryProcessingError != null)
            {
                DirectoryProcessingError(this, new DirectoryProcessingArgs(u.EngineNumber, u.GlobalThreadNumber, "", GetElapsedTimeForEngineNumber(u.EngineNumber)));
            }
        }

        public void OnScriptExecutionError(object sender, ScriptExecutionErrorArgs e)
        {
            if (DirectoryProcessingError != null)
            {
                DirectoryProcessingError(this, new DirectoryProcessingArgs(e.EngineNumber, e.GlobalThreadNumber, e.DirectoryPath, GetElapsedTimeForEngineNumber(e.EngineNumber)));
            }
        }

        public void OnDirectoryProcessingStarted(object sender, DirectoryProcessingArgs d)
        {
            if (DirectoryProcessingStarted != null)
            {
                DirectoryProcessingStarted(this, d);
            }
        }

        public void OnDirectoryProcessingFinished(object sender, DirectoryProcessingArgs d)
        {
            if (DirectoryProcessingFinished != null)
            {
                DirectoryProcessingFinished(this, d);
            }
        }

        public void OnScriptExecutionComplete(object sender, ScriptExecutionEventArgs s)
        {
            //TimeRemainingInformation tri = _Controller.Engines[s.EngineNumber].Engine.Tracker.GetTimeRemainingInformation();

            //WriteOutput(s.EngineNumber, s.GlobalThreadNumber, "Script execution ('" + s.Message +
            //    "') complete on " + _Controller.Engines[s.EngineNumber].Engine.Tracker.EndTime.ToLongDateString() +
            //    " at " + _Controller.Engines[s.EngineNumber].Engine.Tracker.EndTime.ToLongTimeString());
            //WriteOutput(s.EngineNumber, s.GlobalThreadNumber, "Total execution time: " + tri.TotalElapsedTimeLongString);
        }

        public void OnScriptExecutionAborted(object sender, ScriptExecutionEventArgs s)
        {
            if (DirectoryProcessingError != null)
            {
                DirectoryProcessingError(this, new DirectoryProcessingArgs(s.EngineNumber, s.GlobalThreadNumber, "", GetElapsedTimeForEngineNumber(s.EngineNumber)));
            }
        }

        public void OnImageStarted(object sender, ImageEventArgs a)
        {
            if (ImageProcessingStarted != null)
            {
                ImageProcessingStarted(this, a);
            }
        }

        public void OnImageFinished(object sender, ImageFinishedEventArgs a)
        {
            if (a.FirstImageNumber > 0)
            {
                //Controller.Engines[a.EngineNumber].Engine.Tracker.
                // build the estimated total processing time for this engine and add it to the metatracker
                TimeSpan elapsed = DateTime.Now - Controller.Engines[a.EngineNumber].Engine.Tracker.StartTime;
                TimeRemainingInformation tri = Controller.Engines[a.EngineNumber].Engine.Tracker.GetTimeRemainingInformation();
                TimeSpan etAvg = elapsed + tri.EstimatedTimeRemainingAverage;
                //TimeSpan etMax = elapsed + tri.EstimatedTimeRemainingMax;
                TimeSpan etMed = elapsed + tri.EstimatedTimeRemainingMedian;
                _MetaTracker.AddElapsedTime(etAvg);
                //_MetaTracker.AddElapsedTime(etMax);
                _MetaTracker.AddElapsedTime(etMed);
                Console.WriteLine("MetaTracker element count: " + _MetaTracker.Values.Length.ToString());
                if (ImageProcessingFinished != null)
                {
                    ImageProcessingFinished(this, a);
                }
            }
        }

        public void OnEngineControllerExecutionStarted(object sender, EngineManagementEventArgs e)
        {
            //if (e.Message != "")
            //{
            //    WriteOutput(-1, -1, e.Message);
            //}
        }

        public void OnEngineControllerExecutionComplete(object sender, EngineManagementEventArgs e)
        {
            if (EngineControllerExecutionComplete != null)
            {
                EngineControllerExecutionComplete(this, e);
            }
        }

        public void OnEngineExecutionComplete(object sender, DirectoryProcessingArgs d)
        {
            _MetaTracker.AddElapsedTime(d.ElapsedTime);
            if (EngineExecutionComplete != null)
            {
                EngineExecutionComplete(this, d);
            }
        }

        public void OnScriptBeingGenerated(object sender, ThreadedEventArgs t)
        {
            if (ScriptBeingGenerated != null)
            {
                ScriptBeingGenerated(sender, t);
            }
        }

        public void OnScriptGenerated(object sender, CubeScriptGeneratedEventArgs t)
        {
            if (ScriptGenerated != null)
            {
                ScriptGenerated(sender, t);
            }
        }

        public void OnWaitingForProcessingThread(object sender, EngineManagementEventArgs e)
        {
            if (WaitingForProcessingThread != null)
            {
                WaitingForProcessingThread(sender, e);
            }
        }

    }
}
