// BenLincoln.DaVinci - A shared library used by The Mirror's Surface Breaks
// Copyright 2010-2011 Ben Lincoln
// http://www.beneaththewaves.net/
//

// This file is part of BenLincoln.DaVinci.

// BenLincoln.DaVinci is free software: you can redistribute it and/or modify
// it under the terms of version 3 of the GNU General Public License as published by
// the Free Software Foundation.

// BenLincoln.DaVinci is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with BenLincoln.DaVinci (in the file LICENSE.txt).  
// If not, see <http://www.gnu.org/licenses/>.

using System;
using System.Collections;
using System.Collections.Generic;
using System.IO;
using System.Reflection;
using System.Text;
using System.Text.RegularExpressions;
using System.Xml;

namespace BenLincoln.DaVinci
{
    public class TMSBProcessingConfigurationFile : TMSBConfigurationFile
    {
        protected TMSBProcessingConfiguration _ProcessingConfig;
        protected TransformationProfileCollection _TransformationProfiles;

        protected enum ConfigProcessingState
        {
            Normal,
            ColourImageOutputTransformations,
            StatCubesMain,
            StatCube,
            StatCubeMember,
            CalculatedChannelsMain,
            CalculatedChannel,
            CustomGreyscaleChannelMain,
            CustomGreyscaleChannel,
            CustomThreeChannelImagesMain,
            CustomThreeChannelImage,
            ChannelCombinationsMain,
            ChannelCombination,
            DCSSection,
            DCSVariation,
            TintMappingMain,
            TintMapping,
            GradientMappingMain,
            GradientMapping
        }

        #region Properties

        public TMSBProcessingConfiguration ProcessingConfig
        {
            get
            {
                return _ProcessingConfig;
            }
            set
            {
                _ProcessingConfig = value;
            }
        }

        #endregion

        public TMSBProcessingConfigurationFile(string configFileName)
            : this(configFileName, null)
        {
        }

        public TMSBProcessingConfigurationFile(string configFileName, TransformationProfileCollection transformationProfiles)
            : base()
        {
            _ProcessingConfig = new TMSBProcessingConfiguration();
            _TransformationProfiles = transformationProfiles;
            ReadData(configFileName);
        }

        protected override void ReadTypeSpecificData(Stream sourceStream)
        {
            XmlTextReader xReader = new XmlTextReader(sourceStream);
            //xReader.Settings.Schemas.Add(_FileSchema);

            string currentElementName = "";
            ConfigProcessingState currentState = ConfigProcessingState.Normal;
            //CombinedBandConfig statGreyConfig = new CombinedBandConfig();

            StatisticalCubeConfig statCubeConfig = new StatisticalCubeConfig();
            TransformationProfileCollection statCubeConfigInputTransformations = new TransformationProfileCollection();
            TransformationProfileCollection statCubeConfigOutputTransformations = new TransformationProfileCollection();

            CombinedBandConfig statCubeMemberConfig = new CombinedBandConfig();
            TransformationProfileCollection statCubeMemberConfigOutputTransformations = new TransformationProfileCollection();

            //StatisticalFalseColourImageConfig statColourConfig = new StatisticalFalseColourImageConfig();
            CombinedBandConfig channelCombineConfig = new CombinedBandConfig();
            TransformationProfileCollection channelCombineConfigOutputTransformations = new TransformationProfileCollection();

            CustomGreyscaleChannelConfig customGreyscaleConfig = new CustomGreyscaleChannelConfig();
            TransformationProfileCollection customGreyscaleConfigInputTransformations = new TransformationProfileCollection();
            TransformationProfileCollection customGreyscaleConfigOutputTransformations = new TransformationProfileCollection();

            CustomThreeChannelImageConfig customThreeChannelConfig = new CustomThreeChannelImageConfig();
            TransformationProfileCollection customThreeChannelConfigInputTransformations = new TransformationProfileCollection();
            TransformationProfileCollection customThreeChannelConfigOutputTransformations = new TransformationProfileCollection();

            DCSParameterSet dcsVariation = new DCSParameterSet();
            TintMapping newMapping = new TintMapping();
            TintMapping newGradientMapping = new TintMapping();

            CalculatedChannelCombinationConfig newCalculatedChannelCombinationConfig = new CalculatedChannelCombinationConfig();
            TransformationProfileCollection newCalculatedChannelCombinationConfigInputTransformations = new TransformationProfileCollection();
            TransformationProfileCollection newCalculatedChannelCombinationConfigOutputTransformations = new TransformationProfileCollection();

            while (xReader.Read())
            {
                switch (xReader.NodeType)
                {
                    case XmlNodeType.Element:
                        currentElementName = xReader.Name;
                        //Console.WriteLine("XML DEBUG: Begin Element - Current Processing State '" + currentState.ToString() + "' - Element name: '" + currentElementName + "'.");
                        switch (currentState)
                        {
                            case ConfigProcessingState.Normal:
                                switch (currentElementName)
                                {
                                    case "ColourImageOutputTransformations":
                                        currentState = ConfigProcessingState.ColourImageOutputTransformations;
                                        break;
                                    case "StatisticalGreyscaleCubes":
                                        currentState = ConfigProcessingState.StatCubesMain;
                                        break;
                                    case "CustomGreyscaleChannels":
                                        currentState = ConfigProcessingState.CustomGreyscaleChannelMain;
                                        break;
                                    case "CombinedBandChannels":
                                        currentState = ConfigProcessingState.ChannelCombinationsMain;
                                        break;
                                    case "CalculatedChannelSets":
                                        currentState = ConfigProcessingState.CalculatedChannelsMain;
                                        break;
                                    case "CustomThreeChannelImages":
                                        currentState = ConfigProcessingState.CustomThreeChannelImagesMain;
                                        break;
                                    case "DCSVariations":
                                        currentState = ConfigProcessingState.DCSSection;
                                        break;
                                    case "TintMapping":
                                        currentState = ConfigProcessingState.TintMappingMain;
                                        break;
                                    case "GradientMapping":
                                        currentState = ConfigProcessingState.GradientMappingMain;
                                        break;
                                }
                                break;
                            case ConfigProcessingState.StatCubesMain:
                                switch (currentElementName)
                                {
                                    case "StatisticalGreyscaleCube":
                                        statCubeConfig = new StatisticalCubeConfig();
                                        statCubeConfigInputTransformations = new TransformationProfileCollection();
                                        statCubeConfigOutputTransformations = new TransformationProfileCollection();
                                        currentState = ConfigProcessingState.StatCube;
                                        break;
                                }
                                break;
                            case ConfigProcessingState.StatCube:
                                switch (currentElementName)
                                {
                                    case "StatisticalGreyscaleImage":
                                        statCubeMemberConfig = new CombinedBandConfig();
                                        statCubeMemberConfigOutputTransformations = new TransformationProfileCollection();
                                        currentState = ConfigProcessingState.StatCubeMember;
                                        break;
                                }
                                break;
                            case ConfigProcessingState.ChannelCombinationsMain:
                                if (currentElementName == "CombinedChannel")
                                {
                                    channelCombineConfig = new CombinedBandConfig();
                                    channelCombineConfigOutputTransformations = new TransformationProfileCollection();
                                    currentState = ConfigProcessingState.ChannelCombination;
                                }
                                break;
                            case ConfigProcessingState.CalculatedChannelsMain:
                                if (currentElementName == "CalculatedChannelSet")
                                {
                                    newCalculatedChannelCombinationConfig = new CalculatedChannelCombinationConfig();
                                    newCalculatedChannelCombinationConfigInputTransformations = new TransformationProfileCollection();
                                    newCalculatedChannelCombinationConfigOutputTransformations = new TransformationProfileCollection();
                                    currentState = ConfigProcessingState.CalculatedChannel;
                                }
                                break;
                            case ConfigProcessingState.CustomGreyscaleChannelMain:
                                if (currentElementName == "CustomGreyscaleChannel")
                                {
                                    customGreyscaleConfig = new CustomGreyscaleChannelConfig();
                                    customGreyscaleConfigInputTransformations = new TransformationProfileCollection();
                                    customGreyscaleConfigOutputTransformations = new TransformationProfileCollection();
                                    currentState = ConfigProcessingState.CustomGreyscaleChannel;
                                }
                                break;
                            case ConfigProcessingState.CustomThreeChannelImagesMain:
                                if (currentElementName == "CustomThreeChannelImage")
                                {
                                    customThreeChannelConfig = new CustomThreeChannelImageConfig();
                                    customThreeChannelConfigInputTransformations = new TransformationProfileCollection();
                                    customThreeChannelConfigOutputTransformations = new TransformationProfileCollection();
                                    currentState = ConfigProcessingState.CustomThreeChannelImage;
                                }
                                break;
                            case ConfigProcessingState.DCSSection:
                                if (currentElementName == "DCSVariation")
                                {
                                    currentState = ConfigProcessingState.DCSVariation;
                                    dcsVariation = new DCSParameterSet();
                                }
                                break;

                            // beginning of a tint mapping definition
                            case ConfigProcessingState.TintMappingMain:
                                if (currentElementName == "TintedChannel")
                                {
                                    currentState = ConfigProcessingState.TintMapping;
                                    newMapping = new TintMapping();
                                }
                                break;

                            // beginning of the definition of a gradient mapping
                            case ConfigProcessingState.GradientMappingMain:
                                if (currentElementName == "GradientMappedChannel")
                                {
                                    currentState = ConfigProcessingState.GradientMapping;
                                    newGradientMapping = new TintMapping();
                                }
                                break;
                        }
                        break;
                    case XmlNodeType.EndElement:
                        currentElementName = xReader.Name;
                        //Console.WriteLine("XML DEBUG: End Element - Current Processing State '" + currentState.ToString() + "' - Element name: '" + currentElementName + "'.");
                        switch (currentState)
                        {
                            case ConfigProcessingState.ColourImageOutputTransformations:
                                if (currentElementName == "ColourImageOutputTransformations")
                                {
                                    currentState = ConfigProcessingState.Normal;
                                }
                                break;
                            case ConfigProcessingState.StatCubesMain:
                                if (currentElementName == "StatisticalGreyscaleCubes")
                                {
                                    currentState = ConfigProcessingState.Normal;
                                }
                                break;
                            case ConfigProcessingState.StatCube:
                                if (currentElementName == "StatisticalGreyscaleCube")
                                {
                                    statCubeConfig.InputTransformations = statCubeConfigInputTransformations;
                                    statCubeConfig.OutputTransformations = statCubeConfigOutputTransformations;
                                    ProcessingConfig.AddStatisticalCubeConfiguration(statCubeConfig);
                                    currentState = ConfigProcessingState.StatCubesMain;
                                }
                                break;
                            case ConfigProcessingState.StatCubeMember:
                                if (currentElementName == "StatisticalGreyscaleImage")
                                {
                                    statCubeMemberConfig.OutputTransformations = statCubeMemberConfigOutputTransformations;
                                    statCubeConfig.AddStatisticalGreyscaleConfiguration(statCubeMemberConfig);
                                    currentState = ConfigProcessingState.StatCube;
                                }
                                break;

                            case ConfigProcessingState.CustomGreyscaleChannelMain:
                                if (currentElementName == "CustomGreyscaleChannels")
                                {
                                    currentState = ConfigProcessingState.Normal;
                                }
                                break;
                            case ConfigProcessingState.CustomGreyscaleChannel:
                                if (currentElementName == "CustomGreyscaleChannel")
                                {
                                    if (customGreyscaleConfig.IsValidConfiguration())
                                    {
                                        customGreyscaleConfig.InputTransformations = customGreyscaleConfigInputTransformations;
                                        customGreyscaleConfig.OutputTransformations = customGreyscaleConfigOutputTransformations;
                                        ProcessingConfig.AddCustomGreyscaleChannelConfig(customGreyscaleConfig);
                                    }
                                    else
                                    {
                                        throw new ArgumentException("The custom greyscale channel definition '" + customGreyscaleConfig.ChannelName +
                                            "' is not valid. Please double-check the XML configuration.");
                                    }
                                    currentState = ConfigProcessingState.CustomGreyscaleChannelMain;
                                }
                                break;
                            case ConfigProcessingState.CustomThreeChannelImagesMain:
                                if (currentElementName == "CustomThreeChannelImages")
                                {
                                    currentState = ConfigProcessingState.Normal;
                                }
                                break;
                            case ConfigProcessingState.ChannelCombinationsMain:
                                if (currentElementName == "CombinedBandChannels")
                                {
                                    currentState = ConfigProcessingState.Normal;
                                }
                                break;
                            case ConfigProcessingState.CalculatedChannelsMain:
                                if (currentElementName == "CalculatedChannelSets")
                                {
                                    currentState = ConfigProcessingState.Normal;
                                }
                                break;
                            case ConfigProcessingState.CalculatedChannel:
                                if (currentElementName == "CalculatedChannelSet")
                                {
                                    currentState = ConfigProcessingState.CalculatedChannelsMain;
                                    newCalculatedChannelCombinationConfig.InputTransformations = newCalculatedChannelCombinationConfigInputTransformations;
                                    newCalculatedChannelCombinationConfig.OutputTransformations = newCalculatedChannelCombinationConfigOutputTransformations;
                                    ProcessingConfig.AddCalculatedChannelConfiguration(newCalculatedChannelCombinationConfig);
                                }
                                break;
                            case ConfigProcessingState.ChannelCombination:
                                if (currentElementName == "CombinedChannel")
                                {
                                    channelCombineConfig.OutputTransformations = channelCombineConfigOutputTransformations;
                                    ProcessingConfig.AddChannelCombinationMethod(channelCombineConfig);
                                    currentState = ConfigProcessingState.ChannelCombinationsMain;
                                }
                                break;
                            case ConfigProcessingState.CustomThreeChannelImage:
                                if (currentElementName == "CustomThreeChannelImage")
                                {
                                    ProcessingConfig.AddCustomThreeChannelImageConfig(customThreeChannelConfig);
                                    currentState = ConfigProcessingState.CustomThreeChannelImagesMain;
                                }
                                break;
                            case ConfigProcessingState.DCSSection:
                                if (currentElementName == "DCSVariations")
                                {
                                    currentState = ConfigProcessingState.Normal;
                                }
                                break;
                            case ConfigProcessingState.DCSVariation:
                                if (currentElementName == "DCSVariation")
                                {
                                    ProcessingConfig.AddDCSType(dcsVariation);
                                    currentState = ConfigProcessingState.DCSSection;
                                }
                                break;
                            case ConfigProcessingState.TintMappingMain:
                                if (currentElementName == "TintMapping")
                                {
                                    currentState = ConfigProcessingState.Normal;
                                }
                                break;

                            // end of a tint mapping definition
                            case ConfigProcessingState.TintMapping:
                                if (currentElementName == "TintedChannel")
                                {
                                    currentState = ConfigProcessingState.TintMappingMain;
                                    ProcessingConfig.AddTintMapping(newMapping);
                                }
                                break;

                            // end of all gradient-mapping definitions
                            case ConfigProcessingState.GradientMappingMain:
                                if (currentElementName == "GradientMapping")
                                {
                                    currentState = ConfigProcessingState.Normal;
                                }
                                break;

                            // end of a gradient-mapping definition
                            case ConfigProcessingState.GradientMapping:
                                if (currentElementName == "GradientMappedChannel")
                                {
                                    currentState = ConfigProcessingState.GradientMappingMain;
                                    ProcessingConfig.AddGradientMapping(newGradientMapping);
                                }
                                break;
                        }
                        break;
                    case XmlNodeType.Text:
                        string currentText = xReader.Value.Trim();
                        switch (currentState)
                        {
                            case ConfigProcessingState.Normal:
                                switch (currentElementName)
                                {
                                    case "GenerateStatisticalChannels":
                                        ProcessingConfig.GenerateStatisticalChannels = StringToBool(currentText);
                                        break;
                                    case "GenerateCalculatedChannels":
                                        ProcessingConfig.GenerateCalculatedChannels = StringToBool(currentText);
                                        break;
                                    case "GenerateCustomChannels":
                                        ProcessingConfig.GenerateCustomChannels = StringToBool(currentText);
                                        break;
                                    case "GenerateTintedImages":
                                        ProcessingConfig.GenerateTintedImages = StringToBool(currentText);
                                        break;
                                    case "GenerateGradientImages":
                                        ProcessingConfig.GenerateGradientImages = StringToBool(currentText);
                                        break;
                                    case "GenerateThreeChannelImages":
                                        ProcessingConfig.GenerateThreeChannelImages = StringToBool(currentText);
                                        break;
                                    case "GenerateStandardThreeChannelImages":
                                        ProcessingConfig.GenerateStandardThreeChannelImages = StringToBool(currentText);
                                        break;
                                    case "GenerateCustomThreeChannelImages":
                                        ProcessingConfig.GenerateCustomThreeChannelImages = StringToBool(currentText);
                                        break;
                                    case "GenerateCalculatedThreeChannelImages":
                                        ProcessingConfig.GenerateCalculatedThreeChannelImages = StringToBool(currentText);
                                        break;
                                    case "GenerateLumaColourImages":
                                        ProcessingConfig.GenerateLumaColourImages = StringToBool(currentText);
                                        break;
                                    case "GenerateDCSImages":
                                        ProcessingConfig.GenerateDCSImages = StringToBool(currentText);
                                        break;
                                    case "GenerateStatisticalImageThreeChannelPermutations":
                                        ProcessingConfig.GenerateStatisticalImageThreeChannelPermutations = StringToBool(currentText);
                                        break;
                                    case "GenerateStatisticalImageThreeChannelPermutationsUnrestricted":
                                        ProcessingConfig.GenerateStatisticalImageThreeChannelPermutationsUnrestricted = StringToBool(currentText);
                                        break;
                                    case "GenerateLumaColourVariationsFromIndividualBands":
                                        ProcessingConfig.GenerateLumaColourVariationsFromIndividualBands = StringToBool(currentText);
                                        break;
                                    case "GenerateLumaColourVariationsFromCombinedChannels":
                                        ProcessingConfig.GenerateLumaColourVariationsFromCombinedChannels = StringToBool(currentText);
                                        break;
                                    case "GenerateLumaColourVariationsFromStatisticalChannels":
                                        ProcessingConfig.GenerateLumaColourVariationsFromStatisticalChannels = StringToBool(currentText);
                                        break;
                                    case "GenerateLumaColourVariationsFromThreeChannelImages":
                                        ProcessingConfig.GenerateLumaColourVariationsFromThreeChannelImages = StringToBool(currentText);
                                        break;
                                    case "GenerateLumaColourVariationsFromCustomThreeChannelImages":
                                        ProcessingConfig.GenerateLumaColourVariationsFromCustomThreeChannelImages = StringToBool(currentText);
                                        break;
                                    case "GenerateLumaColourVariationsFromTintedImages":
                                        ProcessingConfig.GenerateLumaColourVariationsFromTintedImages = StringToBool(currentText);
                                        break;
                                    case "GenerateLumaColourVariationsFromGradientMappedImages":
                                        ProcessingConfig.GenerateLumaColourVariationsFromGradientMappedImages = StringToBool(currentText);
                                        break;
                                    case "GenerateLumaColourVariationsFromDCSImages":
                                        ProcessingConfig.GenerateLumaColourVariationsFromDCSImages = StringToBool(currentText);
                                        break;
                                    case "GenerateDCSImagesFromIndividualChannels":
                                        ProcessingConfig.GenerateDCSImagesFromIndividualChannels = StringToBool(currentText);
                                        break;
                                    case "GenerateDCSImagesFromCombinedChannels":
                                        ProcessingConfig.GenerateDCSImagesFromCombinedChannels = StringToBool(currentText);
                                        break;
                                    case "GenerateDCSImagesFromCustomThreeChannelImages":
                                        ProcessingConfig.GenerateDCSImagesFromCustomThreeChannelImages = StringToBool(currentText);
                                        break;
                                    case "GenerateDCSEnhancedContrast":
                                        ProcessingConfig.GenerateDCSEnhancedContrast = StringToBool(currentText);
                                        break;
                                }
                                break;
                            case ConfigProcessingState.ColourImageOutputTransformations:
                                switch (currentElementName)
                                {
                                    case "OutputTransformation":
                                        string spName = currentText.Trim();
                                        TransformationProfileCollection.AddTransformationProfileIfItExists(
                                            ref _TransformationProfiles, ref ProcessingConfig.ColourImageOutputTransformations, 
                                            spName, "The cube-wide colour image output transformation list");
                                        break;
                                }
                                break;
                            case ConfigProcessingState.StatCube:
                                switch (currentElementName)
                                {
                                    case "CubeName":
                                        statCubeConfig.CubeName = currentText;
                                        break;
                                    case "WindowSize":
                                        statCubeConfig.WindowSize = int.Parse(currentText);
                                        break;
                                    case "UseAsThreeChannelChannels":
                                        statCubeConfig.UseAsThreeChannelChannels = StringToBool(currentText);
                                        break;
                                    case "InputTransformation":
                                        string spName = currentText.Trim();
                                        TransformationProfileCollection.AddTransformationProfileIfItExists(
                                            ref _TransformationProfiles, ref statCubeConfigInputTransformations,
                                            spName, "The input transformation list for the statistical greyscale cube '" + statCubeConfig.CubeName + "'");
                                        break;
                                    case "OutputTransformation":
                                        string spName2 = currentText.Trim();
                                        TransformationProfileCollection.AddTransformationProfileIfItExists(
                                            ref _TransformationProfiles, ref statCubeConfigOutputTransformations,
                                            spName2, "The output transformation list for the statistical greyscale cube '" + statCubeConfig.CubeName + "'");
                                        break;
                                }
                                break;
                            case ConfigProcessingState.StatCubeMember:
                                switch (currentElementName)
                                {
                                    case "BandCombineMethod":
                                        statCubeMemberConfig.BandCombineMode = GreyscaleChannel.GetBandCombineMethodFromString(currentText);
                                        break;
                                    //case "OutputNormalization":
                                    //    statCubeMemberConfig.OutputNormalization = OutputImage.NormalizationModeFromString(currentText);
                                    //    break;
                                    case "UseAsLuminanceChannel":
                                        statCubeMemberConfig.UseAsLuminanceChannel = StringToBool(currentText);
                                        break;
                                    case "OutputTransformation":
                                        string spName = currentText.Trim();
                                        TransformationProfileCollection.AddTransformationProfileIfItExists(
                                            ref _TransformationProfiles, ref statCubeMemberConfigOutputTransformations,
                                            spName, "The output transformation list specific to one member of the statistical greyscale cube '" + statCubeConfig.CubeName + "'");
                                        break;
                                }
                                break;
                            case ConfigProcessingState.CalculatedChannel:
                                switch (currentElementName)
                                {
                                    case "Name":
                                        newCalculatedChannelCombinationConfig.CombinationConfigName = currentText.Trim();
                                        break;
                                    case "CombinationMethod":
                                        newCalculatedChannelCombinationConfig.CombineMethod = CalculatedChannelCombinationConfig.CalculatedChannelCombinationMethodFromString(currentText);
                                        break;
                                    case "MaxRecursionDepth":
                                        newCalculatedChannelCombinationConfig.MaxRecursionDepth = int.Parse(currentText);
                                        break;
                                    case "GenerateReciprocal":
                                        newCalculatedChannelCombinationConfig.GenerateReciprocal = StringToBool(currentText);
                                        break;
                                    //case "AbsoluteValueOutput":
                                    //    newCalculatedChannelCombinationConfig.AbsoluteValueOutput = StringToBool(currentText);
                                    //    break;
                                    //case "OutputNormalization":
                                    //    newCalculatedChannelCombinationConfig.OutputNormalization = OutputImage.NormalizationModeFromString(currentText);
                                    //    break;
                                    //case "NAAVNormalization":
                                    //    newCalculatedChannelCombinationConfig.NAAVNormalization = OutputImage.NormalizationModeFromString(currentText);
                                    //    break;
                                    case "InputTransformation":
                                        string spName = currentText.Trim();
                                        TransformationProfileCollection.AddTransformationProfileIfItExists(
                                            ref _TransformationProfiles, ref newCalculatedChannelCombinationConfigInputTransformations,
                                            spName, "The input transformation list for the combined channel configuration '" +
                                            newCalculatedChannelCombinationConfig.CombineMethod.ToString() + "'");
                                        break;
                                    case "OutputTransformation":
                                        string spName2 = currentText.Trim();
                                        TransformationProfileCollection.AddTransformationProfileIfItExists(
                                            ref _TransformationProfiles, ref newCalculatedChannelCombinationConfigOutputTransformations,
                                            spName2, "The output transformation list for the combined channel configuration '" + 
                                            newCalculatedChannelCombinationConfig.CombineMethod.ToString() + "'");
                                        break;
                                    case "UseAsLuminanceChannel":
                                        newCalculatedChannelCombinationConfig.UseAsLuminanceChannel = StringToBool(currentText);
                                        break;
                                    case "UseAsStandardThreeChannelChannel":
                                        newCalculatedChannelCombinationConfig.UseAsStandardThreeChannelChannel = StringToBool(currentText);
                                        break;
                                    case "GenerateThreeChannelSimilarMotion":
                                        newCalculatedChannelCombinationConfig.GenerateThreeChannelSimilarMotion = StringToBool(currentText);
                                        break;
                                    case "GenerateThreeChannelObliqueMotion":
                                        newCalculatedChannelCombinationConfig.GenerateThreeChannelObliqueMotion = StringToBool(currentText);
                                        break;
                                    case "GenerateThreeChannelOtherPermutations":
                                        newCalculatedChannelCombinationConfig.GenerateThreeChannelOtherPermutations = StringToBool(currentText);
                                        break;
                                    case "GenerateThreeChannelReciprocal":
                                        newCalculatedChannelCombinationConfig.GenerateThreeChannelReciprocal = StringToBool(currentText);
                                        break;
                                    case "UseThreeChannelPermuatationsInLumaColourImages":
                                        newCalculatedChannelCombinationConfig.UseThreeChannelPermuatationsInLumaColourImages = StringToBool(currentText);
                                        break;
                                }
                                break;

                            case ConfigProcessingState.CustomGreyscaleChannel:
                                switch (currentElementName)
                                {
                                    case "ChannelName":
                                        customGreyscaleConfig.ChannelName = currentText;
                                        break;
                                    case "Type":
                                        customGreyscaleConfig.ChannelType = CustomGreyscaleChannelConfig.CustomGreyscaleChannelTypeFromString(currentText);
                                        break;
                                    case "CategoryName":
                                        customGreyscaleConfig.Category = SpectralBand.BandCategoryFromString(currentText);
                                        break;
                                    case "CustomScriptCode":
                                        customGreyscaleConfig.CustomScriptCode = currentText;
                                        break;
                                    case "BandCombineMethod":
                                        customGreyscaleConfig.BandCombineMode = GreyscaleChannel.GetBandCombineMethodFromString(currentText);
                                        break;
                                    //case "OutputNormalization":
                                    //    customGreyscaleConfig.OutputNormalization = OutputImage.NormalizationModeFromString(currentText);
                                    //    break;
                                    case "UseAsLuminanceChannel":
                                        customGreyscaleConfig.UseAsLuminanceChannel = StringToBool(currentText);
                                        break;
                                    case "UseAsThreeChannelChannel":
                                        customGreyscaleConfig.UseAsThreeChannelChannel = StringToBool(currentText);
                                        break;
                                    case "InputTransformation":
                                        string spName = currentText.Trim();
                                        TransformationProfileCollection.AddTransformationProfileIfItExists(
                                            ref _TransformationProfiles, ref customGreyscaleConfigInputTransformations,
                                            spName, "The input transformation list for the custom greyscale channel '" +
                                            customGreyscaleConfig.ChannelName + "'");
                                        break;
                                    case "OutputTransformation":
                                        string spName2 = currentText.Trim();
                                        TransformationProfileCollection.AddTransformationProfileIfItExists(
                                            ref _TransformationProfiles, ref customGreyscaleConfigOutputTransformations,
                                            spName2, "The output transformation list for the custom greyscale channel '" +
                                            customGreyscaleConfig.ChannelName + "'");
                                        break;
                                }
                                break;
                            case ConfigProcessingState.CustomThreeChannelImage:
                                switch (currentElementName)
                                {
                                    case "UserDefinedName":
                                        customThreeChannelConfig.UserDefinedName = currentText;
                                        break;
                                    case "RedChannelName":
                                        customThreeChannelConfig.RedChannelName = currentText;
                                        break;
                                    case "GreenChannelName":
                                        customThreeChannelConfig.GreenChannelName = currentText;
                                        break;
                                    case "BlueChannelName":
                                        customThreeChannelConfig.BlueChannelName = currentText;
                                        break;
                                    //case "OutputNormalization":
                                    //    customThreeChannelConfig.OutputNormalization = OutputImage.NormalizationModeFromString(currentText);
                                    //    break;
                                    case "InputTransformation":
                                        string spName = currentText.Trim();
                                        TransformationProfileCollection.AddTransformationProfileIfItExists(
                                            ref _TransformationProfiles, ref customThreeChannelConfigInputTransformations,
                                            spName, "The input transformation list for the custom three-channel image '" +
                                            customThreeChannelConfig.UserDefinedName + "'");
                                        break;
                                    case "OutputTransformation":
                                        string spName2 = currentText.Trim();
                                        TransformationProfileCollection.AddTransformationProfileIfItExists(
                                            ref _TransformationProfiles, ref customThreeChannelConfigOutputTransformations,
                                            spName2, "The output transformation list for the custom three-channel image '" +
                                            customThreeChannelConfig.UserDefinedName + "'");
                                        break;
                                    case "IncludeInSecondOrderImages":
                                        customThreeChannelConfig.IncludeInSecondOrderImages = StringToBool(currentText);
                                        break;
                                }
                                break;
                            case ConfigProcessingState.ChannelCombination:
                                switch (currentElementName)
                                {
                                    case "BandCombineMethod":
                                        channelCombineConfig.BandCombineMode = GreyscaleChannel.GetBandCombineMethodFromString(currentText);
                                        break;
                                    //case "OutputNormalization":
                                    //    channelCombineConfig.OutputNormalization = OutputImage.NormalizationModeFromString(currentText);
                                    //    break;
                                    case "UseAsLuminanceChannel":
                                        channelCombineConfig.UseAsLuminanceChannel = StringToBool(currentText);
                                        break;
                                    case "OutputTransformation":
                                        string spName = currentText.Trim();
                                        TransformationProfileCollection.AddTransformationProfileIfItExists(
                                            ref _TransformationProfiles, ref channelCombineConfigOutputTransformations,
                                            spName, "The output transformation list for the channel combination method '" +
                                            channelCombineConfig.BandCombineMode.ToString() + "'");
                                        break;
                                }
                                break;
                            case ConfigProcessingState.DCSVariation:
                                switch (currentElementName)
                                {
                                    case "DCSType":
                                        dcsVariation.Mode = DCSParameterSet.DCSModeFromString(currentText);
                                        break;
                                    case "ColourSpace":
                                        dcsVariation.ColourSpace = DCSParameterSet.DCSColourSpaceFromString(currentText);
                                        break;
                                    case "SubSectionSize":
                                        dcsVariation.RDCSSectionPercent = float.Parse(currentText);
                                        break;
                                    case "Ignore":
                                        dcsVariation.Ignore = float.Parse(currentText);
                                        break;
                                    case "Variance":
                                        dcsVariation.Variance = float.Parse(currentText);
                                        break;
                                    case "UseInStandardLumaColourImages":
                                        dcsVariation.UseInStandardLumaColourImages = StringToBool(currentText);
                                        break;
                                    case "UseInDCSEnhancedContrast":
                                        dcsVariation.UseInDCSEnhancedContrast = StringToBool(currentText);
                                        break;
                                }
                                break;

                            // tint mapping-related data
                            case ConfigProcessingState.TintMapping:
                                switch (currentElementName)
                                {
                                    case "ReferenceType":
                                        newMapping.Channel.ReferenceType = CustomChannelReference.CustomChannelReferenceTypeFromString(currentText);
                                        break;
                                    case "ChannelReference":
                                        newMapping.Channel.ReferenceName = currentText;
                                        break;
                                    case "TintName":
                                        newMapping.TintName = currentText;
                                        break;
                                    case "IncludeInSecondOrderImages":
                                        newMapping.IncludeInSecondOrderImages = StringToBool(currentText);
                                        break;
                                }
                                break;


                            // gradient mappings
                            case ConfigProcessingState.GradientMapping:
                                switch (currentElementName)
                                {
                                    case "ReferenceType":
                                        newGradientMapping.Channel.ReferenceType = CustomChannelReference.CustomChannelReferenceTypeFromString(currentText);
                                        break;
                                    case "ChannelReference":
                                        newGradientMapping.Channel.ReferenceName = currentText;
                                        break;
                                    case "GradientName":
                                        newGradientMapping.TintName = currentText;
                                        break;
                                }
                                break;
                        }
                        break;
                }
            }
        }
    }
}
