// BenLincoln.DaVinci - A shared library used by The Mirror's Surface Breaks
// Copyright 2010-2011 Ben Lincoln
// http://www.beneaththewaves.net/
//

// This file is part of BenLincoln.DaVinci.

// BenLincoln.DaVinci is free software: you can redistribute it and/or modify
// it under the terms of version 3 of the GNU General Public License as published by
// the Free Software Foundation.

// BenLincoln.DaVinci is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with BenLincoln.DaVinci (in the file LICENSE.txt).  
// If not, see <http://www.gnu.org/licenses/>.

using System;
using System.Collections.Generic;
using System.IO;
using System.Text;

namespace BenLincoln.DaVinci
{
    public class ThreeChannelImage : MultiChannelImage
    {
        public ThreeChannelImage()
            : base()
        {
            _NamePrefix = "FC-3C-";
            //_NamePrefix = "FC-";
            _ImageName = "";
            _Channels = new GreyscaleChannel[3];
        }

        public ThreeChannelImage(GreyscaleChannel[] channels)
            : this()
        {
            if (channels.Length != 3)
            {
                throw new ArgumentException("A channel array passed to the constructor for this class must contain exactly three channels.");
            }
            Channels = channels;
        }

        protected override void _GenerateName()
        {
            if ((_Channels[0] == null) || (_Channels[1] == null) || (_Channels[2] == null))
            {
                return;
            }
            if ((_Channels[0].ImageName == "") || (_Channels[1].ImageName == "") || (_Channels[2].ImageName == ""))
            {
                return;
            }
            StringBuilder nameBuilder = new StringBuilder(_NamePrefix);
            nameBuilder.Append(WrapNameComponentIfComplex(_Channels[0].ImageName));
            nameBuilder.Append(_NameComponentSeparator);
            nameBuilder.Append(WrapNameComponentIfComplex(_Channels[1].ImageName));
            nameBuilder.Append(_NameComponentSeparator);
            nameBuilder.Append(WrapNameComponentIfComplex(_Channels[2].ImageName));
            nameBuilder.Append(_NameSuffix);
            _ImageName = nameBuilder.ToString();
        }

        public override string GetCreationScriptCode(string scriptTargetObjectName, bool regenerateExistingFiles)
        {
            string filePath = GetFilePathWithoutExtension();
            StringBuilder builder = new StringBuilder();
            // if the option to recreate existing images is not enabled and the 
            // file already exists, do nothing.
            if ((!regenerateExistingFiles) && AllOutputFilesExist(filePath, OutputFormats))
            {
                builder.Append("# Information: all file-type variations of the three-channel false colour image '");
                builder.Append(_ImageName);
                builder.Append("' already appear to exist. The option to regenerate existing files is not enabled, so the script will not overwrite them.");
                builder.Append(Environment.NewLine);
                return builder.ToString();
            }
            _WillBeCreatedByScript = true;
            builder.Append("# Create the '");
            builder.Append(_ImageName);
            builder.Append("' three-channel false colour image");
            builder.Append(Environment.NewLine);
            if ((regenerateExistingFiles) && (AnyOutputFilesExist(filePath, OutputFormats)))
            {
                builder.Append("# Warning: one or more variations of the target file already exist, but will be overwritten as the option to regenerate existing files is enabled.");
                builder.Append(Environment.NewLine);
                _WillOverwriteExistingData = true;
            }

            builder.Append(GetChannelReadCode("red", 0));
            builder.Append(GetChannelReadCode("green", 1));
            builder.Append(GetChannelReadCode("blue", 2));

            string redChannelName = _TemporaryObjectNamePrefix + "red";
            string greenChannelName = _TemporaryObjectNamePrefix + "green";
            string blueChannelName = _TemporaryObjectNamePrefix + "blue";
            builder.Append(scriptTargetObjectName);
            builder.Append(" = cat(");
            builder.Append(redChannelName);
            builder.Append(", ");
            builder.Append(greenChannelName);
            builder.Append(", axis = z)");
            builder.Append(Environment.NewLine);
            builder.Append(scriptTargetObjectName);
            builder.Append(" = cat(");
            builder.Append(scriptTargetObjectName);
            builder.Append(", ");
            builder.Append(blueChannelName);
            builder.Append(", axis = z)");
            builder.Append(Environment.NewLine);
            builder.Append(GetImageFileWriteCode(scriptTargetObjectName, regenerateExistingFiles));
            // de-allocate objects
            builder.Append(GreyscaleChannel.GetDeallocationScriptCode(redChannelName));
            builder.Append(GreyscaleChannel.GetDeallocationScriptCode(greenChannelName));
            builder.Append(GreyscaleChannel.GetDeallocationScriptCode(blueChannelName));
            builder.Append(GreyscaleChannel.GetDeallocationScriptCode(scriptTargetObjectName));
            return builder.ToString();
        }

        protected string GetChannelReadCode(string channelName, int channelNumber)
        {
            string resultChannelName = _TemporaryObjectNamePrefix + channelName;
            string resultChannelPath = _Channels[channelNumber].GetFilePathWithPrimaryExtension();
            if (resultChannelPath == "")
            {
                throw new ArgumentException("No " + channelName + " channel has been defined for the three-channel image '" + _ImageName + "'.");
            }

            return _Channels[channelNumber].GetChannelFileReadScriptCode("", resultChannelName);

            //StringBuilder builder = new StringBuilder(resultChannelName);
            //builder.Append(" = read(\"");
            //builder.Append(resultChannelPath);
            //builder.Append("\")");
            //builder.Append(Environment.NewLine);
            //return builder.ToString();
        }

        public bool IsRGBImage(string redBandName, string greenBandName, string blueBandName)
        {
    //        if ((componentList.Count == 3) &&
    //(bandsForChannel[0].BandName == _LuminanceChannelRedBandName) &&
    //(bandsForChannel[1].BandName == _LuminanceChannelGreenBandName) &&
    //(bandsForChannel[2].BandName == _LuminanceChannelBlueBandName))
    //        {
    //            isRGB = true;
    //        }
            if ((_Channels[0].SpectralBands.Length == 1)
                && (_Channels[1].SpectralBands.Length == 1)
                && (_Channels[2].SpectralBands.Length == 1)
                && (_Channels[0].SpectralBands[0].BandName == redBandName)
                && (_Channels[1].SpectralBands[0].BandName == greenBandName)
                && (_Channels[2].SpectralBands[0].BandName == blueBandName)
                )
            {
                return true;
            }
            return false;
        }

        public void RemovePrefixIfRGB(string redBandName, string greenBandName, string blueBandName)
        {
            if (IsRGBImage(redBandName, greenBandName, blueBandName))
            {
                NamePrefix = "";
            }
        }

        public override bool ContainsDisallowedChannels()
        {
            foreach (GreyscaleChannel channel in Channels)
            {
                if (!channel.UseAsThreeChannelChannel)
                {
                    return true;
                }
            }
            return false;
        }

    }
}
